# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
from __future__ import print_function
from builtins import range
from cubical.machines.abstract_machine import MasterMachine
from cubical.machines.complex_2x2_machine import Complex2x2Gains
from cubical.machines.complex_W_2x2_machine import ComplexW2x2Gains
import numpy as np
import cubical.kernels

from cubical.tools import logger
from . import machine_types
from cubical.flagging import FL
log = logger.getLogger("jones_chain")


class JonesChain(MasterMachine):
    """
    This class implements a gain machine for an arbitrary chain of Jones matrices. Most of its
    functionality is consistent with a complex 2x2 solver - many of its methods mimic those of the 
    underlying complex 2x2 machines. This specific chain machine implements a chain of robust 2x2 
    solvers
    """

    def __init__(self, label, data_arr, ndir, nmod, times, frequencies, chunk_label, jones_options):
        """
        Initialises a chain of complex 2x2 gain machines.

        Args:
            label (str):
                Label identifying the Jones term.
            data_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed 
                visibilities. 
            ndir (int):
                Number of directions.
            nmod (nmod):
                Number of models.
            times (np.ndarray):
                Times for the data being processed.
            frequencies (np.ndarray):
                Frequencies for the data being processsed.
            jones_options (dict): 
                Dictionary of options pertaining to the chain. 
        """
        from cubical.main import UserInputError
        # This instantiates the number of complex 2x2 elements in our chain. Each element is a
        # gain machine in its own right - the purpose of this machine is to manage these machines
        # and do the relevant fiddling between parameter updates. When combining DD terms with
        # DI terms, we need to be initialise the DI terms using only one direction - we do this with
        # slicing rather than summation as it is slightly faster.
        
        self.jones_terms = []
        self.num_left_di_terms = 0  # how many DI terms are there at the left of the chain
        seen_dd_term = False
        for iterm, term_opts in enumerate(jones_options['chain']):
            jones_class = machine_types.get_machine_class(term_opts['type'])
            if jones_class is None:
                raise UserInputError("unknown Jones class '{}'".format(term_opts['type']))
            if jones_class not in (ComplexW2x2Gains, ComplexW2x2Gains) and term_opts['solvable']:
                raise UserInputError("only complex-2x2 or robust-2x2 terms can be made solvable in a Jones chain")
            term = jones_class(term_opts["label"], data_arr, ndir, nmod, times, frequencies, chunk_label, term_opts)
            self.jones_terms.append(term)
            if term.dd_term:
                seen_dd_term = True
            elif not seen_dd_term:
                self.num_left_di_terms = iterm
            

        MasterMachine.__init__(self, label, data_arr, ndir, nmod, times, frequencies,
                               chunk_label, jones_options)

        self.chain  = cubical.kernels.import_kernel("chain")
        # kernel used for compute_residuals and such
        self.kernel = cubical.kernels.import_kernel("full_W_complex")
        #ComplexW2x2Gains.get_full_kernel(jones_options, diag_gains=self.is_diagonal)

        self.n_dir, self.n_mod = ndir, nmod
        _, self.n_tim, self.n_fre, self.n_ant, self.n_ant, self.n_cor, self.n_cor = data_arr.shape

        self.n_terms = len(self.jones_terms)
        # make list of number of iterations per solvable term
        # If not specified, just use the maxiter setting of each term
        # note that this list is updated as we converge, so make a copy
        term_iters = jones_options['sol']['term-iters']
        if not term_iters:
            self.term_iters = [term.maxiter for term in self.jones_terms if term.solvable]
        elif type(term_iters) is int:
            self.term_iters = [term_iters]
        elif isinstance(term_iters, (list, tuple)):
            self.term_iters = list(term_iters)
        else:
            raise UserInputError("invalid term-iters={} setting".format(term_iters))

        self.solvable = bool(self.term_iters)

        # setup first solvable term in chain
        self.active_index = -1
        self._next_chain_term()

        # this list accumulates the per-term convergence status strings
        self._convergence_states = []
        # True when the last active term has had its convergence status queried
        self._convergence_states_finalized = False

        self.cached_model_arr = self._r = self._m = None

        self.is_robust = True # to identify this machine as the robust solver


    @classmethod
    def determine_diagonality(cls, options):
        """Returns true if the machine class, given the options, represents a diagonal gain"""
        from cubical.main import UserInputError
        diagonal = True
        for term_opts in options['chain']:
            jones_class = machine_types.get_machine_class(term_opts['type'])
            if jones_class is None:
                raise UserInputError("unknown Jones class '{}'".format(term_opts['type']))
            diagonal = diagonal and jones_class.determine_diagonality(term_opts)
        return diagonal

    @classmethod
    def determine_allocators(cls, options):
        kernel = cubical.kernels.import_kernel("full_W_complex")
        #Complex2x2Gains.get_full_kernel(options, diag_gains=cls.determine_diagonality(options))
        return kernel.allocate_vis_array, kernel.allocate_flag_array, kernel.allocate_gain_array

    def precompute_attributes(self, data_arr, model_arr, flags_arr, inv_var_chan):
        """Precomputes various stats before starting a solution"""
        MasterMachine.precompute_attributes(self, data_arr, model_arr, flags_arr, inv_var_chan)
        for term in self.jones_terms:
            if term.solvable:
                term.precompute_attributes(data_arr, model_arr, flags_arr, inv_var_chan)

    def export_solutions(self):
        """ Saves the solutions to a dict of {label: solutions,grids} items. """

        soldict = {}
        # prefix jones label to solution name
        for term in self.jones_terms:
            if term.solvable:
                for label, sol in term.export_solutions().items():
                    soldict["{}:{}".format(term.jones_label, label)] = sol
        soldict['prefixed'] = True

        return soldict

    def importable_solutions(self, grid):
        """ Returns a dictionary of importable solutions for the chain. """

        soldict = {}
        for term in self.jones_terms:
            soldict.update(term.importable_solutions(grid))

        return soldict

    def import_solutions(self, soldict):
        """
        Loads solutions from a dict. This should not be called -- _load_solutions()
        below should rather call import_solutions() on all the chain terms.
        """
        raise RuntimeError("This method cannot be called on a Jones chain. This is a bug.")

    def _load_solutions(self, init_sols, full_grid):
        """
        Helper method invoked by Factory.create_machine() to import existing solutions into machine.

        In the case of a chain, we invoke this method on every member.
        """
        for term in self.jones_terms:
            term._load_solutions(init_sols, full_grid)



    #@profile
    def compute_js(self, obser_arr, model_arr):
        """
        This function computes the (J\ :sup:`H`\J)\ :sup:`-1` and J\ :sup:`H`\R terms of the GN/LM 
        method. This method is more complicated than a more conventional gain machine. The use of
        a chain means there are additional terms which need to be considered when computing the 
        parameter updates. Here each individual gain machines compute weighted variants of the
        above terms since we are using the robust solver.

        Args:
            obser_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                observed visibilities.
            model_arr (np.ndrray): 
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                model visibilities.

        Returns:
            3-element tuple
                
                - J\ :sup:`H`\R (np.ndarray)
                - (J\ :sup:`H`\J)\ :sup:`-1` (np.ndarray)
                - Count of flags raised (int)     
        """     


        n_dir, n_tint, n_fint, n_ant, n_cor, n_cor = self.active_term.gains.shape

        if self.last_active_index!=self.active_index or self.iters==1:
            self.cached_model_arr = cached_model_arr = np.empty_like(model_arr)
            np.copyto(cached_model_arr, model_arr)

            for ind in range(self.n_terms - 1, self.active_index, -1):
                term = self.jones_terms[ind]
                term.apply_gains(cached_model_arr, full2x2=True)

            # collapse direction axis, if current term is non-DD
            if not self.active_term.dd_term and self.n_dir>1:
                self.cached_model_arr = np.empty_like(model_arr[0:1,...])
                np.sum(cached_model_arr, axis=0, keepdims=True, out=self.cached_model_arr)

            self.jh = np.empty_like(self.cached_model_arr)

            jhr_shape = [n_dir if self.active_term.dd_term else 1, self.n_tim, self.n_fre, n_ant, n_cor, n_cor]

            self._jhr = self.active_term.allocate_gain_array(jhr_shape, dtype=obser_arr.dtype)

            jhrint_shape = [n_dir, n_tint, n_fint, n_ant, n_cor, n_cor]

            self._jhrint = self.active_term.allocate_gain_array(jhrint_shape, dtype=self._jhr.dtype)
            self._jhj = np.empty_like(self._jhrint)
            self._jhjinv =  np.empty_like(self._jhrint)

        np.copyto(self.jh, self.cached_model_arr)

        for ind in range(self.active_index, -1, -1):
            term = self.jones_terms[ind]
            self.chain.compute_jh(self.jh, term.gains, *term.gain_intervals)
        
        # for the robust solver you just have to compute the residuals
        if self.iters == 1:
            self.active_term.residuals = np.empty_like(obser_arr) 
            self.active_term.residuals = self.compute_residual(obser_arr, model_arr, self.active_term.residuals, require_full=False)

            if self.last_active_index!=self.active_index:
                self.active_term.v = self.jones_terms[self.last_active_index].v
                self.active_term.not_all_flagged = self.jones_terms[self.last_active_index].not_all_flagged
                self.active_term.new_flags = self.jones_terms[self.last_active_index].new_flags.copy()
                self.active_term.robust_flag_disable = self.jones_terms[self.last_active_index].robust_flag_disable

            if self.not_all_flagged:
                self.active_term.update_weights()
            
            self.residuals = self.active_term.residuals
            
            
        if self.active_term.offdiag_only:
            self.residuals[..., (0, 1), (0, 1)] = 0

        if self.active_term.diag_only:
            self.residuals[..., (0, 1), (1, 0)] = 0 
        
        self._jhr.fill(0)

        #computing jhwr which jhr * the weights
        self.active_term.kernel_robust.compute_jhwr(self.jh, self.residuals, self.weights, self._jhr, 1, 1)

        for ind in range(0, self.active_index, 1):
            term = self.jones_terms[ind]
            g_inv, gh_inv, flag_counts = term.get_inverse_gains()
            self.chain.apply_left_inv_jones(self._jhr, g_inv, *term.gain_intervals)

        self._jhrint.fill(0)
        self.chain.sum_jhr_intervals(self._jhr, self._jhrint, *self.active_term.gain_intervals)

        self._jhj.fill(0)
        self.active_term.kernel_robust.compute_jhwj(self.jh, self.weights, self._jhj, *self.active_term.gain_intervals)

        flag_count = self.kernel.compute_jhjinv(self._jhj, self._jhjinv,
                                                    self.active_term.gflags, self.active_term.eps, FL.ILLCOND)

        return self._jhrint, self._jhjinv, flag_count


    def implement_update(self, jhr, jhjinv):
        return self.active_term.implement_update(jhr, jhjinv)

    def compute_update(self, model_arr, obser_arr):
        """
        This method is expected to compute the parameter update. 
        
        The standard implementation simply calls compute_js() and implement_update(), here we are
        overriding it because we need to update weights as well. 

        Args:
            model_arr (np.ndarray): 
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing 
                model visibilities. 
            
            obser_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed 
                visibilities. 
        """
        
        jhwr, jhwjinv, flag_count = self.compute_js(obser_arr, model_arr)

        self.implement_update(jhwr, jhwjinv)

        if self.not_all_flagged:

            # Computing the weights
            
            self.active_term.residuals = self.compute_residual(obser_arr, model_arr, self.active_term.residuals, require_full=False)

            self.active_term.update_weights()

            self.residuals = self.active_term.residuals

        return flag_count


    def update_weight_flags(self, flags_arr):
        self.active_term.update_weight_flags(flags_arr)

    def robust_flag(self, flags_arr, model_arr, obser_arr, final=False):
        """run an iteration and use the weights to flag high level RFIs"""

        self.active_term.flaground = True
        self.last_active_index = self.active_index
        self.active_term._final_flaground = True if final else False
        
        if final is False:
            self.active_term.v = 5  # Don't start with a low degree of freedom to prevent overflagging
            self.iters = 1

        if final and self.not_all_flagged:
            self.active_term.flag_weights()
        else:
            self.compute_update(model_arr, obser_arr)
            np.copyto(self.active_term.gains, self.active_term.old_gains)

        if self.any_new:
            flags_arr[self.weight_flags[1:-1]] |= FL.MAD
            self.update_equation_counts(flags_arr != 0)

            new_flags = flags_arr & ~(FL.MISSING | FL.PRIOR) != 0
                
            model_arr[:, :, new_flags!=0, :, :] = 0
            obser_arr[   :, new_flags!=0, :, :] = 0

        if final is False:
            self.weights[:, self.new_flags==0, :] = 1
            self.iters = 0
            self.active_term.v = 2

        self.active_term.flaground = False
       

    def accumulate_gains(self, dd=True):
        """
        This function returns the product of all the gains in the chain, at full TF resolution,
        for all directions (dd=True), or just the 0th direction.

        Args:
            dd (bool):
                Accumulate per-direction gains, if available. If False, only the
                first direction is used.

        Returns:
            A tuple of gains,conjugate gains
        """
        ndir = self.n_dir if dd else 1
        gains = self.jones_terms[0].allocate_gain_array([ndir, self.n_tim, self.n_fre, self.n_ant, self.n_cor, self.n_cor],
                                                        self.dtype)
        g0 = self.jones_terms[0]._gainres_to_fullres(self.jones_terms[0].gains, tdim_ind=1)
        if ndir > 1 and g0.shape[0] == 1:
            g0 = g0.reshape(g0.shape[1:])[np.newaxis,...]
        elif ndir == 1 and g0.shape[0] > 1:
            g0 = g0[:1,...]
        gains[:] = g0
        for term in self.jones_terms[1:]:
            term.kernel.right_multiply_gains(gains, term.gains, *term.gain_intervals)

        # compute conjugate gains
        gh = np.empty_like(gains)
        np.conj(gains.transpose(0, 1, 2, 3, 5, 4), gh)

        return gains, gh
        

    def accumulate_inv_gains(self, direction=None):
        """
        This function returns the inverse of the product of all the gains in the chain, at full TF resolution.
        If direction=None, only DI gains are included. Otherwise, also includes DD gains for the given direction.

        Returns:
            A tuple of gains,conjugate gains,flag_count (if flags raised in inversion)
        """
        gains = self.jones_terms[-1].allocate_gain_array([1, self.n_tim, self.n_fre, self.n_ant, self.n_cor, self.n_cor],
                                                  self.dtype)
        init = False
        fc0 = 0
        # flip order of jones terms for inverse
        for term in self.jones_terms[::-1]:
            if direction is not None or not term.dd_term:
                # dirslice will select the specified direction from the GF array
                dirslice = slice(0, 1) if not term.dd_term else slice(direction, direction + 1)
                g, _, fc = term.get_inverse_gains()
                # g = term._gainres_to_fullres(g, tdim_ind=1)
                fc0 += fc
                if init:
                    term.kernel.right_multiply_gains(gains, g[dirslice,...], *term.gain_intervals)
                else:
                    init = True
                    gains[:] = term._gainres_to_fullres(g[dirslice,...], tdim_ind=1)

        # compute conjugate gains
        gh = np.empty_like(gains)
        np.conj(gains.transpose(0, 1, 2, 3, 5, 4), gh)

        return gains, gh, fc



    
    #@profile
    def compute_residual(self, obser_arr, model_arr, resid_arr, require_full=True):
        """
        This function computes the residual. This is the difference between the
        observed data, and the model data with the gains applied to it.

        Args:
            obser_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                observed visibilities.
            model_arr (np.ndrray): 
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                model visibilities.
            resid_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array into which the 
                computed residuals should be placed.

        Returns:
            np.ndarray: 
                Array containing the result of computing D - GMG\ :sup:`H`.
        """

        g, gh = self.accumulate_gains()
        np.copyto(resid_arr, obser_arr)
        kernel = self.kernel if require_full else self.active_term.kernel_robust
        kernel.compute_residual(model_arr, g, gh, resid_arr, 1, 1)

        return resid_arr


    def apply_inv_gains(self, resid_vis, corr_vis=None, full2x2=True, direction=None):
        """
        Applies the inverse of the gain estimates to the observed data matrix.

        Args:
            obser_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                observed visibilities.
            corr_vis (np.ndarray or None, optional): 
                if specified, shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array 
                into which the corrected visibilities should be placed.
            direction (int or None):
                if None, only DI terms will be applied. Otherwise, also applies DD terms for given direction.

        Returns:
            np.ndarray: 
                Array containing the result of G\ :sup:`-1`\DG\ :sup:`-H`.
        """

        if corr_vis is None:
            corr_vis = np.empty_like(resid_vis)

        g, gh, flag_count = self.accumulate_inv_gains(direction=direction)

        self.kernel.compute_corrected(resid_vis, g, gh, corr_vis, 1, 1)

        return corr_vis, flag_count

    def apply_gains(self, vis, full2x2=True):
        """
        Applies the gains to an array at full time-frequency resolution. 

        Args:
            model_arr (np.ndarray):
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing 
                model visibilities.

        Returns:
            np.ndarray:
                Array containing the result of GMG\ :sup:`H`.
        """
        g, gh = self.accumulate_gains()
        self.kernel.apply_gains(vis, g, gh, 1, 1)
        return vis

    def check_convergence(self, min_delta_g):
        """
        Updates the convergence info of the current time-frequency chunk. 

        Args:
            min_delta_g (float):
                Threshold for the minimum change in the gains - convergence criterion.
        """
        return self.active_term.check_convergence(min_delta_g)

    def restrict_solution(self, gains):
        """
        Restricts the solutions by, for example, selecting a reference antenna or taking only the 
        amplitude. 
        """
        return self.active_term.restrict_solution(gains)

    def flag_solutions(self, flags_arr, final=False):
        """ Flags gain solutions."""
        # Per-iteration flagging done on the active term, final flagging is done on all terms.
        if final:
            return any([ term.flag_solutions(flags_arr, True) for term in self.jones_terms if term.solvable ])
        else:
            return self.active_term.flag_solutions(flags_arr, False)

    def num_gain_flags(self, mask=None, final=False):
        if final:
            nfl = [0, 0]
            for term in self.jones_terms:
                n1, n2 = term.num_gain_flags(mask, final=True)
                nfl[0] += int(n1)
                nfl[1] += int(n2)
            return nfl
        else:
            return self.active_term.num_gain_flags(mask)

    def _next_chain_term(self):
        while True:
            if not self.term_iters:
                return False
            previous_term = self.active_term
            # clear converged and stalled properties
            self.active_index = (self.active_index + 1) % self.n_terms
            if self.active_term.solvable:
                self.active_term.maxiter = self.term_iters.pop(0)
                if not self.active_term.maxiter:
                    print("skipping term {}: 0 term iters specified".format(self.active_term.jones_label), file=log(1))
                    continue
                self.active_term.iters = 0
                self._convergence_states_finalized = False
                if previous_term:
                    previous_term.has_converged = previous_term.has_stalled = False
                self.active_term.has_converged = self.active_term.has_stalled = False
                print("activating term {}".format(self.active_term.jones_label), file=log(1))
                return True
            else:
                print("skipping term {}: non-solvable".format(self.active_term.jones_label), file=log(1))


    def next_iteration(self):
        """
        Updates the iteration count on the relevant element of the Jones chain. It will also handle 
        updating the active Jones term. Ultimately, this should handle any complicated 
        convergence/term switching functionality.
        """

        self.last_active_index = self.active_index
        major_step = False

        if self.active_term.has_converged or self.active_term.has_stalled:
            print("term {} {} ({} iters): {}".format(self.active_term.jones_label,
                        "converged" if self.active_term.has_converged else "stalled",
                        self.active_term.iters, self.active_term.final_convergence_status_string), file=log(1))
            self._convergence_states.append(self.active_term.final_convergence_status_string)
            self._convergence_states_finalized = True
            self._next_chain_term()
            major_step = True

        self.active_term.next_iteration()

        return MasterMachine.next_iteration(self)[0], major_step

    def compute_chisq(self, resid_arr, inv_var_chan, require_full=True):
        """Computes chi-square using the active chain term"""
        if require_full:
            return super(JonesChain, self).compute_chisq(resid_arr, inv_var_chan)
        else:
            return self.active_term.compute_chisq(resid_arr, inv_var_chan)

    @property
    def has_valid_solutions(self):
        """Gives corresponding property of the active chain term"""
        return all([term.has_valid_solutions for term in self.jones_terms if term.solvable])

    @property
    def jones_label(self):
        return self.active_term.jones_label

    @property
    def num_converged_solutions(self):
        """Gives corresponding property of the active chain term"""
        return self.active_term.num_converged_solutions

    @property
    def active_term(self):
        return self.jones_terms[self.active_index]

    @property
    def dd_term(self):
        """Gives corresponding property of the active chain term"""
        return any([ term.dd_term for term in self.jones_terms ])

    @property
    def diag_only(self):
        """This property is true if the machine solves using diagonal visibilities only"""
        return all([ term.diag_only for term in self.jones_terms ])

    @property
    def offdiag_only(self):
        """This property is true if the machine solves using off-diagonal visibilities only"""
        return all([ term.offdiag_only for term in self.jones_terms ])

    @property
    def iters(self):
        """Gives corresponding property of the active chain term"""
        return self.active_term.iters

    @iters.setter
    def iters(self, value):
        """Sets corresponding property of the active chain term"""
        self.active_term.iters = value

    @property
    def maxiter(self):
        """Gives corresponding property of the active chain term"""
        return self.active_term.maxiter

    @maxiter.setter
    def maxiter(self, value):
        """Sets corresponding property of the active chain term"""
        self.active_term.maxiter = value

    @property
    def conditioning_status_string(self):
        return "; ".join([term.conditioning_status_string for term in self.jones_terms])

    @property
    def current_convergence_status_string(self):
        """Current status is reported from the active term"""
        return self.active_term.current_convergence_status_string

    @property
    def final_convergence_status_string(self):
        """Final status is reported from all terms"""
        if not self._convergence_states_finalized:
            self._convergence_states.append(self.active_term.final_convergence_status_string)
            self._convergence_states_finalized = True
        return "; ".join(self._convergence_states)

    @property
    def has_converged(self):
        # Chain has converged when term_iters is empty -- since we take off an element each time we converge a term
        return not self.solvable or \
               ( self.active_term.has_converged and not self.term_iters )
        #return np.all([term.has_converged for term in self.jones_terms])

    @property
    def has_stalled(self):
        return self.active_term.has_stalled and not self.term_iters

    @has_stalled.setter   
    def has_stalled(self, value):
        self.active_term.has_stalled = value
        
    @property
    def epsilon(self):
        return self.active_term.epsilon

    @property
    def delta_chi(self):
        return self.active_term.delta_chi

    @property
    def weights(self):
        return self.active_term.weights

    @property
    def new_flags(self):
        return self.active_term.new_flags
    
    @property
    def weight_flags(self):
        return self.active_term.weight_flags

    @property
    def any_new(self):
        return self.active_term.any_new

    @property
    def not_all_flagged(self):
        return self.active_term.not_all_flagged

    @property
    def robust_flag_disable(self):
        return self.active_term.robust_flag_disable

    @property
    def save_weights(self):
        """Gives corresponding property of the active chain term"""
        return any([ term.save_weights for term in self.jones_terms ])

    @property
    def robust_flag_weights(self):
        """Gives corresponding property of the active chain term"""
        return any([ term.robust_flag_weights for term in self.jones_terms ])
    
    
    
    class Factory(MasterMachine.Factory):
        """
        Note that a ChainMachine Factory expects a list of jones options (one dict per Jones term), not a single dict.
        """
        def __init__(self, machine_cls, grid, double_precision, apply_only, global_options, jones_options):
            # manufacture dict of "Jones options" for the outer chain
            opts = dict(label="chain", solvable=not apply_only, sol=global_options['sol'], chain=jones_options)
            if apply_only:
                for term in jones_options:
                    term['solvable'] = False
            self.chain_options = jones_options
            MasterMachine.Factory.__init__(self, machine_cls, grid, double_precision, apply_only,
                                           global_options, opts)

        def init_solutions(self):
            for opts in self.chain_options:
                label = opts["label"]
                jones_class = machine_types.get_machine_class(opts['type'])
                self._init_solutions(label,
                                     self.make_filename(opts["xfer-from"], label) or
                                     self.make_filename(opts["load-from"], label),
                                     bool(opts["xfer-from"]),
                                     self.solvable and opts["solvable"] and self.make_filename(opts["save-to"], label),
                                     jones_class.exportable_solutions(),
                                     dd_term=opts["dd-term"])

        def determine_allocators(self):
            return self.machine_class.determine_allocators(self.jones_options)





# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Kernels for 2x2 complex visibilities with diagonal gains. Functions require output arrays to be 
provided. Common dimensions of arrays are:

+----------------+------+
| Dimension      | Size |
+================+======+
| Direction      |   d  |
+----------------+------+
| Model          |   m  |
+----------------+------+
| Time           |   t  |
+----------------+------+
| Time Intervals |   ti |
+----------------+------+
| Frequency      |   f  |
+----------------+------+
| Freq Intervals |   fi |
+----------------+------+
| Antenna        |   a  |
+----------------+------+
| Correlation    |   c  |
+----------------+------+

"""
from builtins import range

import numpy as np
from numba import jit, prange

import cubical.kernels
from cubical.kernels import generics
from cubical.kernels import full_complex

use_parallel = True if cubical.kernels.num_omp_threads > 1 else False
use_cache = cubical.kernels.use_cache

# Allocators same as for generic full kernel
allocate_vis_array = full_complex.allocate_vis_array
allocate_gain_array = full_complex.allocate_gain_array
allocate_flag_array = full_complex.allocate_flag_array
allocate_param_array = full_complex.allocate_param_array

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_residual(m, g, gh, r, t_int, f_int):
    """
    Given the model, gains, and their conjugates, computes the residual. Residual has full time and
    frequency resolution - solution intervals are used to correctly associate the gains with the
    model.

    Args:
        m (np.complex64 or np.complex128):
            Typed memoryview of model array with dimensions (d, m, t, f, a, a, c, c).
        g (np.complex64 or np.complex128):
            Typed memoryview of gain array with dimension (d, ti, fi, a, c, c).
        gh (np.complex64 or np.complex128):
            Typed memoryview of conjugate gain array with dimensions (d, ti, fi, a, c, c).
        r (np.complex64 or np.complex128):
            Typed memoryview of residual array with dimensions (m, t, f, a, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """
    n_dir = m.shape[0]
    n_mod = m.shape[1]
    n_tim = m.shape[2]
    n_fre = m.shape[3]
    n_ant = m.shape[4]
    g_dir = g.shape[0]

    bls = np.array([[i,j] for i in range(n_ant) for j in range(i+1, n_ant)], dtype=np.int32)
    n_bl = bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])
    broadcast_dirs = np.array([d%g_dir for d in range(n_dir)])

    for ibl in prange(n_bl):
        aa, ab = bls[ibl][0], bls[ibl][1]  
        for i in range(n_mod):
            for t in range(n_tim):
                bt = broadcast_times[t]
                for f in range(n_fre):
                    bf = broadcast_freqs[f]
                    for d in range(n_dir):
                        bd = broadcast_dirs[d]
                    
                        g00 = g[bd,bt,bf,aa,0,0]
                        g11 = g[bd,bt,bf,aa,1,1]

                        m00 = m[d,i,t,f,aa,ab,0,0]
                        m01 = m[d,i,t,f,aa,ab,0,1]
                        m10 = m[d,i,t,f,aa,ab,1,0]
                        m11 = m[d,i,t,f,aa,ab,1,1]

                        gh00 = gh[bd,bt,bf,ab,0,0]
                        gh11 = gh[bd,bt,bf,ab,1,1]

                        r[i,t,f,aa,ab,0,0] -= g00*m00*gh00
                        r[i,t,f,aa,ab,0,1] -= g00*m01*gh11
                        r[i,t,f,aa,ab,1,0] -= g11*m10*gh00
                        r[i,t,f,aa,ab,1,1] -= g11*m11*gh11

                    r[i,t,f,ab,aa,0,0] = r[i,t,f,aa,ab,0,0].conjugate()
                    r[i,t,f,ab,aa,1,0] = r[i,t,f,aa,ab,0,1].conjugate()
                    r[i,t,f,ab,aa,0,1] = r[i,t,f,aa,ab,1,0].conjugate()
                    r[i,t,f,ab,aa,1,1] = r[i,t,f,aa,ab,1,1].conjugate()

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_jh(m, g, jh, t_int, f_int):
    """
    Given the model and gains, computes the non-zero elements of J\ :sup:`H`. J\ :sup:`H` has full
    time and frequency resolution - solution intervals are used to correctly associate the gains
    with the model. The result here contains the useful elements of J\ :sup:`H` but does not look
    like the analytic solution.

    Args:
        m (np.complex64 or np.complex128):
            Typed memoryview of model array with dimensions (d, m, t, f, a, a, c, c).
        g (np.complex64 or np.complex128):
            Typed memoryview of gain array with dimension (d, ti, fi, a, c, c).
        jh (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H` array with dimensions (d, m, t, f, a, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """
    n_dir = m.shape[0]
    n_mod = m.shape[1]
    n_tim = m.shape[2]
    n_fre = m.shape[3]
    n_ant = m.shape[4]

    g_dir = g.shape[0]

    all_bls = np.array([[i,j] for i in range(n_ant) for j in range(n_ant) if i!=j])
    n_bl = all_bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])
    broadcast_dirs = np.array([d%g_dir for d in range(n_dir)])

    for ibl in prange(n_bl):
        aa, ab = all_bls[ibl,0], all_bls[ibl,1]
        for i in range(n_mod):
            for t in range(n_tim):
                bt = broadcast_times[t]
                for f in range(n_fre):
                    bf = broadcast_freqs[f]
                    for d in range(n_dir):
                        bd = broadcast_dirs[d]

                        g00 = g[bd,bt,bf,aa,0,0]
                        g11 = g[bd,bt,bf,aa,1,1]

                        m00 = m[d,i,t,f,aa,ab,0,0]
                        m01 = m[d,i,t,f,aa,ab,0,1]
                        m10 = m[d,i,t,f,aa,ab,1,0]
                        m11 = m[d,i,t,f,aa,ab,1,1]
                        
                        jh[d,i,t,f,aa,ab,0,0] = g00*m00
                        jh[d,i,t,f,aa,ab,0,1] = g00*m01
                        jh[d,i,t,f,aa,ab,1,0] = g11*m10
                        jh[d,i,t,f,aa,ab,1,1] = g11*m11

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_update(jhr, jhjinv, upd):
    """
    Given J\ :sup:`H`\R and (J\ :sup:`H`\J)\ :sup:`-1`, computes the gain update. The dimensions of
    the input should already be consistent, making this operation simple.

    Args:
        jhr (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H`\R array with dimensions (d, ti, fi, a, c, c).
        jhjinv (np.complex64 or np.complex128):
            Typed memoryview of (J\ :sup:`H`\J)\ :sup:`-1` array with dimensions
            (d, ti, fi, a, c, c).
        upd (np.complex64 or np.complex128):
            Typed memoryview of gain update array with dimensions (d, ti, fi, a, c, c).
    """

    n_dir = jhr.shape[0]
    n_tim = jhr.shape[1]
    n_fre = jhr.shape[2]
    n_ant = jhr.shape[3]

    for aa in prange(n_ant):
        for t in range(n_tim):
            for f in range(n_fre):
                for d in range(n_dir):

                    jhr00 = jhr[d,t,f,aa,0,0]
                    jhr01 = jhr[d,t,f,aa,0,1]
                    jhr10 = jhr[d,t,f,aa,1,0]
                    jhr11 = jhr[d,t,f,aa,1,1]

                    jhjinv00 = jhjinv[d,t,f,aa,0,0]
                    jhjinv01 = jhjinv[d,t,f,aa,0,1]
                    jhjinv10 = jhjinv[d,t,f,aa,1,0]
                    jhjinv11 = jhjinv[d,t,f,aa,1,1]
                    
                    upd[d,t,f,aa,0,0] = (jhr00*jhjinv00 + jhr01*jhjinv10) 
                    upd[d,t,f,aa,0,1] = 0
                    upd[d,t,f,aa,1,0] = 0
                    upd[d,t,f,aa,1,1] = (jhr10*jhjinv01 + jhr11*jhjinv11)

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_corrected(o, g, gh, corr, t_int, f_int):
    """
    Given the observed visbilities, inverse gains, and their conjugates, computes the corrected
    visibilitites.

    Args:
        o (np.complex64 or np.complex128):
            Typed memoryview of observed visibility array with dimensions (t, f, a, a, c, c).
        g (np.complex64 or np.complex128):
            Typed memoryview of inverse gain array with dimensions (d, ti, fi, a, c, c).
        gh (np.complex64 or np.complex128):
            Typed memoryview of inverse conjugate gain array with dimensions (d, ti, fi, a, c, c).
        corr (np.complex64 or np.complex128):
            Typed memoryview of corrected data array with dimensions (t, f, a, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = g.shape[0]
    n_tim = o.shape[0]
    n_fre = o.shape[1]
    n_ant = o.shape[2]

    bls = np.array([[i,j] for i in range(n_ant) for j in range(i, n_ant)], dtype=np.int32)
    n_bl = bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])

    for ibl in prange(n_bl):
        aa, ab = bls[ibl][0], bls[ibl][1]
        for t in range(n_tim):
            bt = broadcast_times[t]
            for f in range(n_fre):
                bf = broadcast_freqs[f]

                g00 = g[0,bt,bf,aa,0,0]
                g11 = g[0,bt,bf,aa,1,1]

                o00 = o[t,f,aa,ab,0,0]
                o01 = o[t,f,aa,ab,0,1]
                o10 = o[t,f,aa,ab,1,0]
                o11 = o[t,f,aa,ab,1,1]

                gh00 = gh[0,bt,bf,ab,0,0]
                gh11 = gh[0,bt,bf,ab,1,1]

                corr[t,f,aa,ab,0,0] = g00*o00*gh00
                corr[t,f,aa,ab,0,1] = g00*o01*gh11
                corr[t,f,aa,ab,1,0] = g11*o10*gh00
                corr[t,f,aa,ab,1,1] = g11*o11*gh11

                corr[t,f,ab,aa,0,0] = corr[t,f,aa,ab,0,0].conjugate()
                corr[t,f,ab,aa,1,0] = corr[t,f,aa,ab,0,1].conjugate()
                corr[t,f,ab,aa,0,1] = corr[t,f,aa,ab,1,0].conjugate()
                corr[t,f,ab,aa,1,1] = corr[t,f,aa,ab,1,1].conjugate()

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def apply_gains(m, g, gh, t_int, f_int):
    """
    Applies the gains and their conjugates to the model array. This operation is performed in place
    - be wary of losing the original array. The result has full time and frequency resolution -
    solution intervals are used to correctly associate the gains with the model.

    Args:
        m (np.complex64 or np.complex128):
            Typed memoryview of model visibility array with dimensions (d, m , t, f, a, a, c, c).
        g (np.complex64 or np.complex128):
            Typed memoryview of gain array with dimensions (d, ti, fi, a, c, c).
        gh (np.complex64 or np.complex128):
            Typed memoryview of conjugate gain array with dimensions (d, ti, fi, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = m.shape[0]
    n_mod = m.shape[1]
    n_tim = m.shape[2]
    n_fre = m.shape[3]
    n_ant = m.shape[4]

    g_dir = g.shape[0]

    bls = np.array([[i,j] for i in range(n_ant) for j in range(i+1, n_ant)], dtype=np.int32)
    n_bl = bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])
    broadcast_dirs = np.array([d%g_dir for d in range(n_dir)])

    for ibl in prange(n_bl):
        aa, ab = bls[ibl][0], bls[ibl][1]  
        for i in range(n_mod):
            for t in range(n_tim):
                bt = broadcast_times[t]
                for f in range(n_fre):
                    bf = broadcast_freqs[f]
                    for d in range(n_dir):
                        bd = broadcast_dirs[d]
                        
                        g00 = g[bd,bt,bf,aa,0,0]
                        g11 = g[bd,bt,bf,aa,1,1]

                        m00 = m[d,i,t,f,aa,ab,0,0]
                        m01 = m[d,i,t,f,aa,ab,0,1]
                        m10 = m[d,i,t,f,aa,ab,1,0]
                        m11 = m[d,i,t,f,aa,ab,1,1]

                        gh00 = gh[bd,bt,bf,ab,0,0]
                        gh11 = gh[bd,bt,bf,ab,1,1]

                        m[d,i,t,f,aa,ab,0,0] = g00*m00*gh00
                        m[d,i,t,f,aa,ab,0,1] = g00*m01*gh11
                        m[d,i,t,f,aa,ab,1,0] = g11*m10*gh00
                        m[d,i,t,f,aa,ab,1,1] = g11*m11*gh11

                        m[d,i,t,f,ab,aa,0,0] = m[d,i,t,f,aa,ab,0,0].conjugate()
                        m[d,i,t,f,ab,aa,1,0] = m[d,i,t,f,aa,ab,0,1].conjugate()
                        m[d,i,t,f,ab,aa,0,1] = m[d,i,t,f,aa,ab,1,0].conjugate()
                        m[d,i,t,f,ab,aa,1,1] = m[d,i,t,f,aa,ab,1,1].conjugate()

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def right_multiply_gains(g, g_next, t_int, f_int):
    """
    Multiples two gain terms in place. Result has full time and frequency resolution 
    even if g_next does does not. For use in Jones chain.

    NOTE: THIS MAY BE INCORRECT/BACKWARDS.

    Args:
        g (np.complex64 or np.complex128):
            Typed memoryview of gain array with dimensions (d, t, f, a, c, c).
        g_next (np.complex64 or np.complex128):
            Typed memoryview of gain array with dimensions (d, ti, fi, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = g.shape[0]
    n_tim = g.shape[1]
    n_fre = g.shape[2]
    n_ant = g.shape[3]

    g_dir = g_next.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])
    broadcast_dirs = np.array([d%g_dir for d in range(n_dir)])

    for aa in prange(n_ant):
        for t in range(n_tim):
            bt = broadcast_times[t]
            for f in range(n_fre):
                bf = broadcast_freqs[f]
                for d in range(n_dir):
                    bd = broadcast_dirs[d]

                    g00 = g[d,t,f,aa,0,0]
                    g11 = g[d,t,f,aa,1,1]

                    g_next00 = g_next[bd,bt,bf,aa,0,0]
                    g_next01 = g_next[bd,bt,bf,aa,0,1]
                    g_next10 = g_next[bd,bt,bf,aa,1,0]
                    g_next11 = g_next[bd,bt,bf,aa,1,1]

                    g[d,t,f,aa,0,0] *= g_next00
                    g[d,t,f,aa,0,1] *= g_next11
                    g[d,t,f,aa,1,0] *= g_next00
                    g[d,t,f,aa,1,1] *= g_next11

# Map compute_jhr method to a generic complex case.
compute_jhr = full_complex.compute_jhr

# Map compute_jhj method to a generic complex case.
compute_jhj = full_complex.compute_jhj

# Map the J^H.J inversion method to a generic inversion.
compute_jhjinv = generics.compute_2x2_inverse

# Map inversion to generic diagonal inverse.
invert_gains = generics.compute_diag_inverse



from __future__ import print_function
from builtins import range
import multiprocessing, os, sys, traceback
import concurrent.futures as cf
import re
import numba
import numpy as np

import cubical.kernels
from cubical.tools import logger
from cubical import solver


log = logger.getLogger("main")

worker_process_properties = dict(MainProcess={})

# number of worker processes that will be run
global num_workers
num_workers = 0
global num_threads
num_threads = 0

def _setup_workers_and_threads(force_serial, ncpu, nworkers, nthreads, montblanc_threads, max_workers):
    """
    Internal helper -- determines how many workers and threads to allocate, based on defaults specified. See discussion in
    https://github.com/ratt-ru/CubiCal/pull/171#issuecomment-388334586

    Returns:
        Tuple of parallel,nworkers,nthreads
    """
    # generate montblanc status string for use in log reports below
    if montblanc_threads is None:
        montblanc = ""
    elif montblanc_threads:
        montblanc = ", --montblanc-threads {}".format(montblanc_threads)
    else:
        montblanc = ", unlimited Montblanc threads"
    if force_serial:
        cubical.kernels.num_omp_threads = nthreads
        if nthreads:
            nthreads = max(nthreads, montblanc_threads or 1)
            print("forcing single-process mode, {} OMP and/or Montblanc threads".format(nthreads), file=log(0, "blue"))
        elif montblanc_threads:
            nthreads = montblanc_threads
            print("forcing single-process mode, single thread{}".format(montblanc), file=log(0, "blue"))
        return False, 0, nthreads
    if nworkers and nthreads:
        print("multi-process mode: --dist-nworker {} (+1), --dist-nthread {}{}".format(nworkers, nthreads, montblanc), file=log(0, "blue"))
        return True, nworkers, nthreads
    if ncpu:
        cores = ncpu - (montblanc_threads or 1)
        if not nworkers and not nthreads:
            if max_workers:
                cores = min(cores, max_workers)
            print("multi-process mode: {}+1 workers, single thread{}".format(cores, montblanc), file=log(0, "blue"))
            return True, cores, 1
        if nworkers:
            if max_workers:
                nworkers = min(nworkers, max_workers)
            nthreads = max(1, cores // nworkers)
            print("multi-process mode: --dist-nworker {} (+1), {} OMP threads{}".format(nworkers, nthreads, montblanc), file=log(0, "blue"))
            return True, nworkers, nthreads
        if nthreads:
            nworkers = max(1, cores // nthreads)
            if max_workers:
                nworkers = min(nworkers, max_workers)
            print("multi-process mode: {}+1 workers, --dist-nthread {}{}".format(nworkers, nthreads, montblanc), file=log(0, "blue"))
            return True, nworkers, nthreads
    else:  # ncpu not set, and nworkers/nthreads not both set
        if nworkers:
            if max_workers:
                nworkers = min(nworkers, max_workers)
            print("multi-process mode: --dist-nworker {} (+1), single thread{}".format(nworkers, montblanc), file=log(0, "blue"))
            return True, nworkers, 1
        if nthreads:
            print("single-process mode: --dist-thread {}{}".format(nthreads, montblanc), file=log(0, "blue"))
            return False, 0, nthreads
        print("single-process, single-thread mode{}".format(montblanc), file=log(0, "blue"))
        return False, 0, 0
    raise RuntimeError("can't be here -- this is a bug!")


def setup_parallelism(ncpu, nworker, nthread, force_serial, affinity, io_affinity, main_affinity, use_montblanc, montblanc_threads, max_workers):
    """
    Sets up parallelism, affinities and other properties of worker processes.
    
    Args:
        ncpu (int):
            max number of cores to use
        nworker (int):
            Number of workers to run (excluding the I/O worker). If 0, determine automatically.
        nthread (int):
            Number of threads to run per worker. If 0, determine automatically.
        force_serial (bool):
            If True, force serial mode, disabling worker parallelism (threads can still be used)
        affinity (int or str or None):
            If None or empty string, all CPU affinity setting is disabled.
            An "N:M" string specifies allocating starting with core N, stepping by M. An int N specifies
            starting with core N, stepping by 1.
        io_affinity (int):
            If set, enables affinity setting on the I/O worker, and allocates that number of cores to it
            (which includes montblanc threads), overriding the montblanc_threads setting.
        main_affinity (bool or str):
            If set, allocates a separate core to the main process and sets its affinity. If set to "io", pins
            the main process to the same cores as allocated to the I/O worker. If not set, the main process
            is not pinned to any core.
        use_montblanc (bool):
            True if montblanc is being used to predict the model
        montblanc_threads:
            Number of threads to allocate to montblanc. If 0, it will be unlimited.

    Returns:
        True if parallel mode is invoked (i.e. workers are to be launched)
    """
    global num_workers
    global num_threads
    def set_numba_threading(nthread):
        try:
            numba.config.THREADING_LAYER = "safe"
            @numba.njit(parallel=True)
            def foo(a, b):
                return a + b
            foo(np.arange(5), np.arange(5))
            return nthread
        except:
            numba.config.THREADING_LAYER = "workqueue"
            print("Cannot use TBB threading (check your installation). Reverting to workqueue parallelization and dropping to 1 thread per worker in solvers", file=log(0, "red"))
            return 1
    nthread = set_numba_threading(nthread)
    parallel, num_workers, nthread = _setup_workers_and_threads(force_serial, ncpu, nworker, nthread,
                                                                montblanc_threads if use_montblanc else None, max_workers)  
    num_threads = nthread 
    # in serial mode, simply set the Montblanc and/or worker thread count, and return
    if not parallel:
        if nthread:
            cubical.kernels.num_omp_threads = nthread
            numba.set_num_threads(int(nthread))
            os.environ["OMP_NUM_THREADS"] = os.environ["OMP_THREAD_LIMIT"] = str(nthread)
        return False

    # TODO: check actual number of cores on the system, and throw an error if affinity settings exceed this
    # (at the moment we just get an error down the line from libgomp and/or taskset)

    # child processes will inherit this
    cubical.kernels.num_omp_threads = nthread

    numba.set_num_threads(int(nthread))

    # parse affinity argument
    if affinity != "" and affinity is not None:
        if type(affinity) is int or re.match("^[\d+]$", affinity):
            core = int(affinity)
            corestep = 1
        elif re.match("^(\d+):(\d+)$", affinity):
            core, corestep = list(map(int, affinity.split(":")))
        else:
            raise ValueError("invalid affinity setting '{}'".format(affinity))
    else:
        core = corestep = affinity = None

    # first, the I/O process
    props = worker_process_properties["Process-1"] = worker_process_properties["ForkProcess-1"] = \
        dict(label="io", environ={})

    # allocate cores to I/O process, if asked to pin it
    
    # for now, since we can't figure out this tensorflow affinity shit
    if affinity is not None and io_affinity and use_montblanc:
        io_affinity = None
        print("Montblanc currently does not support CPU affinity settings: ignoring --dist-pin-io", file=log(0,"red"))
        
    if affinity is not None and io_affinity:
        num_io_cores = montblanc_threads if use_montblanc else 1
        io_cores = list(range(core,core+num_io_cores*corestep,corestep))
        core = core + num_io_cores * corestep
        # if Montblanc is in use, affinity controlled by GOMP setting, else by taskset
        if use_montblanc:
            props["environ"]["GOMP_CPU_AFFINITY"] = " ".join(map(str,io_cores))
            props["environ"]["OMP_NUM_THREADS"] = props["environ"]["OMP_THREAD_LIMIT"] = str(montblanc_threads)
        else:
            props["taskset"] = ",".join(io_cores)
    # else just restrict Montblanc threads, if asked to
    else:
        io_cores = []
        if use_montblanc and montblanc_threads:
            props["environ"]["OMP_NUM_THREADS"] = props["environ"]["OMP_THREAD_LIMIT"] = str(montblanc_threads)

    # are we asked to pin the main process?
    if affinity is not None and main_affinity:
        if main_affinity == "io":
            if io_affinity:
                worker_process_properties["MainProcess"]["taskset"] = str(io_cores[0])
        else:
            worker_process_properties["MainProcess"]["taskset"] = str(core)
            core = core + corestep

    # create entries for subprocesses, and allocate cores
    for icpu in range(1, num_workers + 1):
        name = "Process-{}".format(icpu + 1)
        name2 = "ForkProcess-{}".format(icpu + 1)
        props = worker_process_properties[name] = worker_process_properties[name2] = \
            dict(label="x%02d" % icpu, num_omp_threads=nthread, environ={})
        if affinity is not None:
            props["taskset"] = str(core)
            # if OMP is in use, set affinities via gomp
            if nthread:
                worker_cores = list(range(core, core + nthread * corestep, corestep))
                core += nthread * corestep
                props["environ"]["GOMP_CPU_AFFINITY"] = " ".join(map(str, worker_cores))
            else:
                core += corestep
    return True


def run_process_loop(ms, _tile_list, load_model, single_chunk, solver_type, solver_opts, debug_opts, out_opts):

    """
    Runs the main loop. If debugging is set, or single_chunk mode is on, forces serial mode.
    Otherwise selects serial or parallel depending on previous call to setup_parallelism().

    Args:
        ms:
        nworker:
        nthread:
        load_model:
        single_chunk:
        debugging:
        solver_type:
        solver_opts:
        debug_opts:
        out_opts:
    Returns:
        Stats dictionary
    """

    # if worker processes are launched, this global is inherited and will be accessed
    # by the I/O worker
    global tile_list
    tile_list = _tile_list
    
    if num_workers:
        return _run_multi_process_loop(ms, load_model, solver_type, solver_opts, debug_opts, out_opts)
    else:
        return _run_single_process_loop(ms, load_model, single_chunk, solver_type, solver_opts, debug_opts, out_opts)


def _run_multi_process_loop(ms, load_model, solver_type, solver_opts, debug_opts, out_opts):
    """
    Runs the main loop in multiprocessing mode.
    
    Args:
        ms: 
        nworker: 
        nthread: 
        load_model: 
        single_chunk: 
        solver_type: 
        solver_opts: 
        debug_opts: 
        out_opts:
    Returns:
        Stats dictionary
    """
    global tile_list

    # this accumulates SolverStats objects from each chunk, for summarizing later
    stats_dict = {}

    def reap_children():
        pid, status, _ = os.wait3(os.WNOHANG)
        if pid:
            print("child process {} exited with status {}. This is a bug, or an out-of-memory condition.".format(pid, status), file=log(0,"red"))
            print("This error is not recoverable: the main process will now commit ritual harakiri.", file=log(0,"red"))
            os._exit(1)
            raise RuntimeError("child process {} exited with status {}".format(pid, status))

    with cf.ProcessPoolExecutor(max_workers=num_workers) as executor, \
            cf.ProcessPoolExecutor(max_workers=1) as io_executor:
        ms.flush()
        # this will be a dict of tile number: future loading that tile
        io_futures = {}
        # schedule I/O job to load tile 0
        io_futures[0] = io_executor.submit(_io_handler, load=0, save=None, load_model=load_model, out_opts=out_opts)
        # all I/O will be done by the I/O thread, so we need to close the MS in the main process
        # and reopen it afterwards
        ms.close()

        # now that the I/O child is forked, init main process properties (affinity etc.)
        # (Montblanc is finicky about affinities apparently, so we don't do it before)
        _init_worker(main=True)

        for itile, tile in enumerate(tile_list):
            # wait for I/O job on current tile to finish
            print("waiting for I/O on {}".format(tile.label), file=log(0))
            # have a timeout so that if a child process dies, we at least find out
            done = False
            while not done:
                reap_children()
                done, not_done = cf.wait([io_futures[itile]], timeout=10)

            # check if result was successful
            if not io_futures[itile].result():
                raise RuntimeError("I/O job on {} failed".format(tile.label))
            del io_futures[itile]

            # immediately schedule I/O job to save previous/load next tile
            load_next = itile + 1 if itile < len(tile_list) - 1 else None
            save_prev = itile - 1 if itile else None
            if load_next is not None or save_prev is not None:
                io_futures[itile + 1] = io_executor.submit(_io_handler, load=load_next,
                                                           save=save_prev, load_model=load_model)

            # submit solver jobs
            solver_futures = {}

            print("submitting solver jobs for {}".format(tile.label), file=log(0))

            for key in tile.get_chunk_keys():
                solver_futures[executor.submit(solver.run_solver, solver_type, itile, key, solver_opts, debug_opts)] = key
                print("submitted solver job for chunk {}".format(key), file=log(3))

            # wait for solvers to finish
            while solver_futures:
                reap_children()
                done, not_done = cf.wait(list(solver_futures.keys()), timeout=1)
                for future in done:
                    key = solver_futures[future]
                    stats = future.result()
                    stats_dict[tile.get_chunk_indices(key)] = stats
                    print("handled result of chunk {}".format(key), file=log(3))
                    del solver_futures[future]

            print("finished processing {}".format(tile.label), file=log(0))

        # ok, at this stage we've iterated over all the tiles, but there's an outstanding
        # I/O job saving the second-to-last tile (which was submitted with itile+1), and the last tile was
        # never saved, so submit a job for that (also to close the MS), and wait
        io_futures[-1] = io_executor.submit(_io_handler, load=None, save=-1, finalize=True)
        done = False
        while not done:
            reap_children()
            done, not_done = cf.wait([io_futures[-1]], timeout=10)

        # get flagcounts from result of job
        ms.update_flag_counts(io_futures[-1].result()['flagcounts'])

        # and reopen the MS again
        ms.reopen()

        return stats_dict


def _run_single_process_loop(ms, load_model, single_chunk, solver_type, solver_opts, debug_opts, out_opts):
    """
    Runs the main loop in single-CPU mode.

    Args:
        ms: 
        nworker: 
        nthread: 
        load_model: 
        single_chunk: 
        solver_type: 
        solver_opts: 
        debug_opts: 
        out_opts: output options from GD
    Returns:
        Stats dictionary
    """
    global tile_list

    stats_dict = {}

    for itile, tile in enumerate(tile_list):
        tile.load(load_model=load_model)
        processed = False
        for key in tile.get_chunk_keys():
            if not single_chunk or key == single_chunk:
                processed = True
                stats = solver.run_solver(solver_type, itile, key, solver_opts, debug_opts)
                stats_dict[tile.get_chunk_indices(key)] = stats

        if processed:
            only_save = ["output", "model", "weight", "flag", "bitflag"] if out_opts is None or out_opts["apply-solver-flags"] else \
                        ["output", "model", "weight"]
            tile.save(final=tile is tile_list[-1], only_save=only_save)
            for sd in tile.iterate_solution_chunks():
                solver.gm_factory.save_solutions(sd)
                solver.ifrgain_machine.accumulate(sd)
        else:
            print("  single-chunk {} not in this tile, skipping it.".format(single_chunk), file=log(0))
        tile.release(final=(itile == len(tile_list) - 1))
        # break out after single chunk is processed
        if processed and single_chunk:
            print("single-chunk {} was processed in this tile. Will now finish".format(single_chunk), file=log(0, "red"))
            break
    solver.ifrgain_machine.save()
    solver.gm_factory.close()

    return stats_dict

# this will be set to True after _init_worker is called()
_worker_initialized = None

def _init_worker(main=False):
    """
    Called inside every worker process first thing, to initialize its properties
    """
    global _worker_initialized
    # for the main process, do not set the singleton guard, as child processes will inherit it
    if not _worker_initialized:
        if not main:
            _worker_initialized = True

        name = multiprocessing.current_process().name
        if name not in worker_process_properties:
            print("WARNING: unrecognized worker process name '{}'. " \
                                "Please inform the developers.".format(name), file=log(0, "red"))
            return
        props = worker_process_properties[name]

        label = props.get("label")
        if label is not None:
            logger.set_subprocess_label(label)

        taskset = props.get("taskset")
        if taskset is not None:
            print("pid {}, setting CPU affinity to {} with taskset".format(os.getpid(), taskset), file=log(1,"blue"))
            os.system("taskset -pc {} {} >/dev/null".format(taskset, os.getpid()))

        environ = props.get("environ")
        if environ:
            os.environ.update(environ)
            for key, value in environ.items():
                print("setting {}={}".format(key, value), file=log(1,"blue"))

        num_omp_threads = props.get("num_omp_threads")
        if num_omp_threads is not None:
            print("enabling {} OMP threads".format(num_omp_threads), file=log(1,"blue"))
            import cubical.kernels
            cubical.kernels.num_omp_threads = num_omp_threads


def _io_handler(save=None, load=None, load_model=True, finalize=False, out_opts=None):
    """
    Handles disk reads and writes for the multiprocessing case.

    Args:
        save (None or int, optional):
            If specified, corresponds to index of Tile to save.
        load (None or int, optional):
            If specified, corresponds to index of Tile to load.
        load_model (bool, optional):
            If specified, loads model column from measurement set.
        finalize (bool, optional):
            If True, this is the last tile. Call the unlock method on the handler, and finalize everything else.
        out_opts (dict):
            Output options from GD
    Returns:
        bool:
            True if load/save was successful.
    """
    global tile_list
    try:
        _init_worker()
        result = {'success': True}
        if save is not None:
            tile = tile_list[save]
            itile = list(range(len(tile_list)))[save]
            print("saving {}".format(tile.label), file=log(0, "blue"))
            only_save = ["output", "model", "weight", "flag", "bitflag"] if out_opts is None or out_opts["apply-solver-flags"] else \
                        ["output", "model", "weight"]
            tile.save(final=finalize, only_save=only_save)
            for sd in tile.iterate_solution_chunks():
                solver.gm_factory.save_solutions(sd)
                solver.ifrgain_machine.accumulate(sd)
            if finalize:
                solver.ifrgain_machine.save()
                solver.gm_factory.close()
                result['flagcounts'] = tile.dh.flagcounts
            tile.release(final=(itile == len(tile_list) - 1))
        if load is not None:
            tile = tile_list[load]
            print("loading {}".format(tile.label), file=log(0, "blue"))
            tile.load(load_model=load_model)
        print("I/O job(s) complete", file=log(0, "blue"))
        return result
    except Exception as exc:
        print("I/O handler for load {} save {} failed with exception: {}".format(load, save, exc), file=log(0, "red"))
        print(traceback.format_exc(), file=log)
        raise




# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Cython kernels for the tf plane machine. Functions require output arrays to be
provided. Common dimensions of arrays are:

+----------------+------+
| Dimension      | Size |
+================+======+
| Direction      |   d  |
+----------------+------+
| Model          |   m  |
+----------------+------+
| Time           |   t  |
+----------------+------+
| Time Intervals |   ti |
+----------------+------+
| Frequency      |   f  |
+----------------+------+
| Freq Intervals |   fi |
+----------------+------+
| Antenna        |   a  |
+----------------+------+
| Correlation    |   c  |
+----------------+------+

"""
from builtins import range

import numpy as np
from numba import jit, prange

import cubical.kernels
from cubical.kernels import generics
from cubical.kernels import diag_complex
from cubical.kernels import phase_only
from cubical.kernels import tf_plane

use_parallel = True if cubical.kernels.num_omp_threads > 1 else False
use_cache = cubical.kernels.use_cache

# Allocators same as for generic full kernel.
allocate_vis_array = tf_plane.allocate_vis_array
allocate_gain_array = tf_plane.allocate_gain_array
allocate_flag_array = tf_plane.allocate_flag_array
allocate_param_array = tf_plane.allocate_param_array

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_jhj(tmp_jhj, jhj, ts, fs, t_int, f_int):
    """
    Given the intermediary J\ :sup:`H`\J and channel frequencies, computes the diagonal entries of
    J\ :sup:`H`\J. J\ :sup:`H`\J is computed over intervals. This is an approximation of the
    Hessian. In the phase-only frequency slope case, this approximation does not depend on the
    gains, therefore it does not vary with iteration. The addition of the block dimension is
    required due to the block-wise diagonal nature of J\ :sup:`H`\J.

    Args:
        tmp_jhj (np.float32 or np.float64):
            Typed memoryview of intermiadiary J\ :sup:`H`\J array with dimensions
            (d, t, f, a, c, c).
        jhj (np.float32 or np.float64):
            Typed memoryview of J\ :sup:`H`\J array with dimensions (d, ti, fi, a, b, c, c).
        fs (np.float32 or np.float64):
            Typed memoryview of channel frequencies with dimension (f).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = tmp_jhj.shape[0]
    n_tim = tmp_jhj.shape[1]
    n_fre = tmp_jhj.shape[2]
    n_ant = tmp_jhj.shape[3]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])

    for aa in prange(n_ant):
        for t in range(n_tim):
            bt = broadcast_times[t]
            for f in range(n_fre):
                bf = broadcast_freqs[f]
                for d in range(n_dir):
                    for c in range(2):

                        tmp_jhjcc = tmp_jhj[d,t,f,aa,c,c]

                        tt = ts[t]

                        jhj[d,bt,bf,aa,0,c,c] += tmp_jhjcc
                        jhj[d,bt,bf,aa,1,c,c] += tt*tmp_jhjcc
                        jhj[d,bt,bf,aa,2,c,c] += tt*tt*tmp_jhjcc

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_jhjinv(jhj, jhjinv, eps):
    """
    Given J\ :sup:`H`\J (or an array with similar dimensions), computes its inverse.

    Args:
        jhj (np.float32 or np.float64):
            Typed memoryview of J\ :sup:`H`\J array with dimensions (d, ti, fi, a, b, c, c).
        jhjinv (np.float32 or np.float64):
            Typed memoryview of (J\ :sup:`H`\J)\ :sup:`-1` array with dimensions
            (d, ti, fi, a, b, c, c).
        eps (float):
            Threshold beneath which the denominator is regarded as too small for inversion.
    """

    n_dir = jhj.shape[0]
    n_tim = jhj.shape[1]
    n_fre = jhj.shape[2]
    n_ant = jhj.shape[3]

    for aa in prange(n_ant):
        for t in range(n_tim):
            for f in range(n_fre):
                for d in range(n_dir):
                    for c in range(2):
                        jhj0cc = jhj[d,t,f,aa,0,c,c]
                        jhj1cc = jhj[d,t,f,aa,1,c,c]
                        jhj2cc = jhj[d,t,f,aa,2,c,c]

                        det = jhj0cc*jhj2cc - jhj1cc*jhj1cc
                        
                        if det<eps:

                            jhjinv[d,t,f,aa,0,c,c] = 0
                            jhjinv[d,t,f,aa,1,c,c] = 0
                            jhjinv[d,t,f,aa,2,c,c] = 0

                        else:

                            det = 1/det

                            jhjinv[d,t,f,aa,0,c,c] =    det*jhj2cc
                            jhjinv[d,t,f,aa,1,c,c] =   -det*jhj1cc
                            jhjinv[d,t,f,aa,2,c,c] =    det*jhj0cc

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_jhr(tmp_jhr, jhr, ts, fs, t_int, f_int):
    """
    Given the intermediary J\ :sup:`H`\R and channel frequencies, computes J\ :sup:`H`\R.
    J\ :sup:`H`\R is computed over intervals. The addition of the block dimension is
    required due to the structure of J\ :sup:`H`.

    Args:
        tmp_jhr (np.float32 or np.float64):
            Typed memoryview of intermediary J\ :sup:`H`\R array with dimensions
            (d, t, f, a, c, c).
        jhr (np.float32 or np.float64):
            Typed memoryview of J\ :sup:`H`\J array with dimensions (d, ti, fi, a, b, c, c).
        fs (np.float32 or np.float64):
            Typed memoryview of channel frequencies with dimension (f).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = tmp_jhr.shape[0]
    n_tim = tmp_jhr.shape[1]
    n_fre = tmp_jhr.shape[2]
    n_ant = tmp_jhr.shape[3]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])

    for aa in prange(n_ant):
        for t in range(n_tim):
            bt = broadcast_times[t]
            for f in range(n_fre):
                bf = broadcast_freqs[f]
                for d in range(n_dir):
                    for c in range(2):
                        tmp_jhrcc = tmp_jhr[d,t,f,aa,c,c]

                        jhr[d,bt,bf,aa,0,c,c] +=       tmp_jhrcc
                        jhr[d,bt,bf,aa,1,c,c] += ts[t]*tmp_jhrcc

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_update(jhr, jhj, upd):
    """
    Given J\ :sup:`H`\R and (J\ :sup:`H`\J)\ :sup:`-1`, computes the gain update. The dimensions of
    the input should already be consistent, making this operation simple. These arrays are real
    valued.

    Args:
        jhr (np.float32 or np.float64):
            Typed memoryview of J\ :sup:`H`\R array with dimensions (d, ti, fi, a, b, c, c).
        jhjinv (np.float32 or np.float64):
            Typed memoryview of (J\ :sup:`H`\J)\ :sup:`-1` array with dimensions
            (d, ti, fi, a, b, c, c).
        upd (np.float32 or np.float64):
            Typed memoryview of gain update array with dimensions (d, ti, fi, a, b, c, c).
    """

    n_dir = jhr.shape[0]
    n_tim = jhr.shape[1]
    n_fre = jhr.shape[2]
    n_ant = jhr.shape[3]

    for aa in prange(n_ant):
        for t in range(n_tim):
            for f in range(n_fre):
                for d in range(n_dir):
                    for c in range(2):

                        jhj0cc = jhj[d,t,f,aa,0,c,c]
                        jhj1cc = jhj[d,t,f,aa,1,c,c]
                        jhj2cc = jhj[d,t,f,aa,2,c,c]

                        jhr0cc = jhr[d,t,f,aa,0,c,c]
                        jhr1cc = jhr[d,t,f,aa,1,c,c]

                        upd[d,t,f,aa,0,c,c] = jhj0cc*jhr0cc + jhj1cc*jhr1cc
                        upd[d,t,f,aa,1,c,c] = jhj1cc*jhr0cc + jhj2cc*jhr1cc

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def construct_gains(param, g, ts, fs, t_int, f_int):
    """
    Given the real-valued parameters of the gains, computes the complex gains.

    Args:
        param (np.float32 or np.float64):
            Typed memoryview of parameter array with dimensions (d, ti, fi, a, b, c, c).
        g (np.complex64 or np.complex128):
            Typed memoryview of gain array with dimensions (d, t, f, a, c, c).
        fs (np.float32 or np.float64):
            Typed memoryview of channel frequencies with dimension (f).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = g.shape[0]
    n_tim = g.shape[1]
    n_fre = g.shape[2]
    n_ant = g.shape[3]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])

    for aa in prange(n_ant):
        for t in range(n_tim):
            bt = broadcast_times[t]
            for f in range(n_fre):
                bf = broadcast_freqs[f]
                for d in range(n_dir):
                    for c in range(2):
                        p0cc = param[d,bt,bf,aa,0,c,c]
                        p1cc = param[d,bt,bf,aa,1,c,c]
                        p2cc = param[d,bt,bf,aa,2,c,c]

                        g[d,t,f,aa,c,c] = np.exp(1j*(p0cc + ts[t]*p1cc))
                               

# Cherry-pick other methods from standard kernels

# inner_jhj is just a phase J^H.J, with intervals of 1,1
compute_inner_jhj = lambda m,jhj1: phase_only.compute_jhj(m, jhj1, 1, 1)

# inner_jhr is just a phase J^H.R with intervals of 1,1
compute_inner_jhr = lambda jh,gh,r,jhr1: phase_only.compute_jhr(jh, gh, r, jhr1, 1, 1)

# J^H computed using diagonal gains
compute_jh = diag_complex.compute_jh

# residuals computed assuming diagonal gains
compute_residual = diag_complex.compute_residual

# corrected visibilities computed assuming diagonal gains
compute_corrected = diag_complex.compute_corrected

# gains applied as diagonal
apply_gains = diag_complex.apply_gains

# gains inverted as diagonal
invert_gains = generics.compute_diag_inverse



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Cython kernels for the 2x2 complex gain machine. Functions require output arrays to be
provided. Common dimensions of arrays are:

+----------------+------+
| Dimension      | Size |
+================+======+
| Direction      |   d  |
+----------------+------+
| Model          |   m  |
+----------------+------+
| Time           |   t  |
+----------------+------+
| Time Intervals |   ti |
+----------------+------+
| Frequency      |   f  |
+----------------+------+
| Freq Intervals |   fi |
+----------------+------+
| Antenna        |   a  |
+----------------+------+
| Correlation    |   c  |
+----------------+------+

"""
from builtins import range

import numpy as np
from numba import jit, prange

import cubical.kernels
from cubical.kernels import generics

use_parallel = True if cubical.kernels.num_omp_threads > 1 else False
use_cache = cubical.kernels.use_cache

# defines memory layout of model-like arrays (axis layout is NDxNMxNTxNFxNAxNAxNCxNC)
_model_axis_layout = [4,5,1,2,3,0,6,7]    # layout is AAMTFD

# defines memory layout of gain-like arrays  (axis layout is NDxNTxNFxNAxNCxNC)
_gain_axis_layout = [3,1,2,0,4,5]   # layout is ATFD

# defines memory layout of flag-like arrays  (axis layout is NTxNFxNAxNA)
_flag_axis_layout = [2,3,0,1]      # layout is AATF

def allocate_vis_array(shape, dtype, zeros=False):
    """Allocates a visibility array of the desired shape, laid out in preferred order"""
    return cubical.kernels.allocate_reordered_array(shape, dtype, _model_axis_layout, zeros=zeros)

def allocate_gain_array(shape, dtype, zeros=False):
    """Allocates a gain array of the desired shape, laid out in preferred order"""
    return cubical.kernels.allocate_reordered_array(shape, dtype, _gain_axis_layout, zeros=zeros)

def allocate_flag_array(shape, dtype, zeros=False):
    """Allocates a flag array of the desired shape, laid out in preferred order"""
    return cubical.kernels.allocate_reordered_array(shape, dtype, _flag_axis_layout, zeros=zeros)

allocate_param_array = allocate_gain_array

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_residual(m, g, gh, r, t_int, f_int):
    """
    Given the model, gains, and their conjugates, computes the residual. Residual has full time and
    frequency resolution - solution intervals are used to correctly associate the gains with the
    model.

    Args:
        m (np.complex64 or np.complex128):
            Typed memoryview of model array with dimensions (d, m, t, f, a, a, c, c).
        g (np.complex64 or np.complex128):
            Typed memoryview of gain array with dimension (d, ti, fi, a, c, c).
        gh (np.complex64 or np.complex128):
            Typed memoryview of conjugate gain array with dimensions (d, ti, fi, a, c, c).
        r (np.complex64 or np.complex128):
            Typed memoryview of residual array with dimensions (m, t, f, a, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """
    n_dir = m.shape[0]
    n_mod = m.shape[1]
    n_tim = m.shape[2]
    n_fre = m.shape[3]
    n_ant = m.shape[4]
    g_dir = g.shape[0]

    bls = np.array([[i,j] for i in range(n_ant) for j in range(i+1, n_ant)], dtype=np.int32)
    n_bl = bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])
    broadcast_dirs = np.array([d%g_dir for d in range(n_dir)])

    for ibl in prange(n_bl):
        aa, ab = bls[ibl][0], bls[ibl][1]  
        for i in range(n_mod):
            for t in range(n_tim):
                bt = broadcast_times[t]
                for f in range(n_fre):
                    bf = broadcast_freqs[f]
                    for d in range(n_dir):
                        bd = broadcast_dirs[d]
                    
                        g00 = g[bd,bt,bf,aa,0,0]
                        g01 = g[bd,bt,bf,aa,0,1]
                        g10 = g[bd,bt,bf,aa,1,0]
                        g11 = g[bd,bt,bf,aa,1,1]

                        m00 = m[d,i,t,f,aa,ab,0,0]
                        m01 = m[d,i,t,f,aa,ab,0,1]
                        m10 = m[d,i,t,f,aa,ab,1,0]
                        m11 = m[d,i,t,f,aa,ab,1,1]

                        gh00 = gh[bd,bt,bf,ab,0,0]
                        gh01 = gh[bd,bt,bf,ab,0,1]
                        gh10 = gh[bd,bt,bf,ab,1,0]
                        gh11 = gh[bd,bt,bf,ab,1,1]

                        r[i,t,f,aa,ab,0,0] -= ((g00*m00 + g01*m10)*gh00 + (g00*m01 + g01*m11)*gh10)
                        r[i,t,f,aa,ab,0,1] -= ((g00*m00 + g01*m10)*gh01 + (g00*m01 + g01*m11)*gh11)
                        r[i,t,f,aa,ab,1,0] -= ((g10*m00 + g11*m10)*gh00 + (g10*m01 + g11*m11)*gh10)
                        r[i,t,f,aa,ab,1,1] -= ((g10*m00 + g11*m10)*gh01 + (g10*m01 + g11*m11)*gh11)

                    r[i,t,f,ab,aa,0,0] = r[i,t,f,aa,ab,0,0].conjugate()
                    r[i,t,f,ab,aa,1,0] = r[i,t,f,aa,ab,0,1].conjugate()
                    r[i,t,f,ab,aa,0,1] = r[i,t,f,aa,ab,1,0].conjugate()
                    r[i,t,f,ab,aa,1,1] = r[i,t,f,aa,ab,1,1].conjugate()     

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_jh(m, g, jh, t_int, f_int):
    """
    Given the model and gains, computes the non-zero elements of J\ :sup:`H`. J\ :sup:`H` has full
    time and frequency resolution - solution intervals are used to correctly associate the gains
    with the model. The result here contains the useful elements of J\ :sup:`H` but does not look
    like the analytic solution.

    Args:
        m (np.complex64 or np.complex128):
            Typed memoryview of model array with dimensions (d, m, t, f, a, a, c, c).
        g (np.complex64 or np.complex128):
            Typed memoryview of gain array with dimension (d, ti, fi, a, c, c).
        jh (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H` array with dimensions (d, m, t, f, a, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """
    n_dir = m.shape[0]
    n_mod = m.shape[1]
    n_tim = m.shape[2]
    n_fre = m.shape[3]
    n_ant = m.shape[4]

    g_dir = g.shape[0]

    all_bls = np.array([[i,j] for i in range(n_ant) for j in range(n_ant) if i!=j])
    n_bl = all_bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])
    broadcast_dirs = np.array([d%g_dir for d in range(n_dir)])

    for ibl in prange(n_bl):
        aa, ab = all_bls[ibl,0], all_bls[ibl,1]
        for i in range(n_mod):
            for t in range(n_tim):
                bt = broadcast_times[t]
                for f in range(n_fre):
                    bf = broadcast_freqs[f]
                    for d in range(n_dir):
                        bd = broadcast_dirs[d]

                        g00 = g[bd,bt,bf,aa,0,0]
                        g01 = g[bd,bt,bf,aa,0,1]
                        g10 = g[bd,bt,bf,aa,1,0]
                        g11 = g[bd,bt,bf,aa,1,1]

                        m00 = m[d,i,t,f,aa,ab,0,0]
                        m01 = m[d,i,t,f,aa,ab,0,1]
                        m10 = m[d,i,t,f,aa,ab,1,0]
                        m11 = m[d,i,t,f,aa,ab,1,1]
                        
                        jh[d,i,t,f,aa,ab,0,0] = (g00*m00 + g01*m10) 
                        jh[d,i,t,f,aa,ab,0,1] = (g00*m01 + g01*m11) 
                        jh[d,i,t,f,aa,ab,1,0] = (g10*m00 + g11*m10)
                        jh[d,i,t,f,aa,ab,1,1] = (g10*m01 + g11*m11)

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_jhr(jh, r, jhr, t_int, f_int):
    """
    Given J\ :sup:`H` and the residual (or observed data, in special cases), computes J\ :sup:`H`\R.
    J\ :sup:`H`\R is computed over intervals.

    Args:
        jh (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H` array with dimensions (d, m, t, f, a, a, c, c).
        r (np.complex64 or np.complex128):
            Typed memoryview of residual array with dimensions (m, t, f, a, a, c, c).
        jhr (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H`\R array with dimensions (d, ti, fi, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = jh.shape[0]
    n_mod = jh.shape[1]
    n_tim = jh.shape[2]
    n_fre = jh.shape[3]
    n_ant = jh.shape[4]

    all_bls = np.array([[i,j] for i in range(n_ant) for j in range(n_ant) if i!=j], dtype=np.int32)
    n_bl = all_bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])

    for ibl in prange(n_bl):
        aa, ab = all_bls[ibl][0], all_bls[ibl][1]
        for i in range(n_mod):
            for t in range(n_tim):
                bt = broadcast_times[t]
                for f in range(n_fre):
                    bf = broadcast_freqs[f]
                    for d in range(n_dir):
                        
                        r00 = r[i,t,f,aa,ab,0,0]
                        r01 = r[i,t,f,aa,ab,0,1]
                        r10 = r[i,t,f,aa,ab,1,0]
                        r11 = r[i,t,f,aa,ab,1,1]

                        jhh00 = jh[d,i,t,f,ab,aa,0,0]
                        jhh01 = jh[d,i,t,f,ab,aa,0,1]
                        jhh10 = jh[d,i,t,f,ab,aa,1,0]
                        jhh11 = jh[d,i,t,f,ab,aa,1,1]

                        jhr[d,bt,bf,aa,0,0] += (r00*jhh00 + r01*jhh10) 
                        jhr[d,bt,bf,aa,0,1] += (r00*jhh01 + r01*jhh11)
                        jhr[d,bt,bf,aa,1,0] += (r10*jhh00 + r11*jhh10)
                        jhr[d,bt,bf,aa,1,1] += (r10*jhh01 + r11*jhh11) 

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_jhj(jh, jhj, t_int, f_int):
    """
    Given J\ :sup:`H` ,computes the diagonal entries of J\ :sup:`H`\J. J\ :sup:`H`\J is computed
    over intervals. This is an approximation of the Hessian.

    Args:
        jh (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H` array with dimensions (d, m, t, f, a, a, c, c).
        jhj (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H`\J array with dimensions (d, ti, fi, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = jh.shape[0]
    n_mod = jh.shape[1]
    n_tim = jh.shape[2]
    n_fre = jh.shape[3]
    n_ant = jh.shape[4]

    all_bls = np.array([[i,j] for i in range(n_ant) for j in range(n_ant) if i!=j], dtype=np.int32)
    n_bl = all_bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])

    for ibl in prange(n_bl):
        aa, ab = all_bls[ibl][0], all_bls[ibl][1]
        for i in range(n_mod):
            for t in range(n_tim):
                bt = broadcast_times[t]
                for f in range(n_fre):
                    bf = broadcast_freqs[f]
                    for d in range(n_dir):

                        j00 = jh[d,i,t,f,ab,aa,0,0]
                        j01 = jh[d,i,t,f,ab,aa,0,1]
                        j10 = jh[d,i,t,f,ab,aa,1,0] 
                        j11 = jh[d,i,t,f,ab,aa,1,1]

                        jh00 = jh[d,i,t,f,ab,aa,0,0].conjugate()
                        jh01 = jh[d,i,t,f,ab,aa,1,0].conjugate()
                        jh10 = jh[d,i,t,f,ab,aa,0,1].conjugate() 
                        jh11 = jh[d,i,t,f,ab,aa,1,1].conjugate()

                        jhj[d,bt,bf,aa,0,0] += (jh00*j00 + jh01*j10)
                        jhj[d,bt,bf,aa,0,1] += (jh00*j01 + jh01*j11)
                        jhj[d,bt,bf,aa,1,0] += (jh10*j00 + jh11*j10)
                        jhj[d,bt,bf,aa,1,1] += (jh10*j01 + jh11*j11)

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_update(jhr, jhjinv, upd):
    """
    Given J\ :sup:`H`\R and (J\ :sup:`H`\J)\ :sup:`-1`, computes the gain update. The dimensions of
    the input should already be consistent, making this operation simple.

    Args:
        jhr (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H`\R array with dimensions (d, ti, fi, a, c, c).
        jhjinv (np.complex64 or np.complex128):
            Typed memoryview of (J\ :sup:`H`\J)\ :sup:`-1` array with dimensions
            (d, ti, fi, a, c, c).
        upd (np.complex64 or np.complex128):
            Typed memoryview of gain update array with dimensions (d, ti, fi, a, c, c).
    """

    n_dir = jhr.shape[0]
    n_tim = jhr.shape[1]
    n_fre = jhr.shape[2]
    n_ant = jhr.shape[3]

    for aa in prange(n_ant):
        for t in range(n_tim):
            for f in range(n_fre):
                for d in range(n_dir):

                    jhr00 = jhr[d,t,f,aa,0,0]
                    jhr01 = jhr[d,t,f,aa,0,1]
                    jhr10 = jhr[d,t,f,aa,1,0]
                    jhr11 = jhr[d,t,f,aa,1,1]

                    jhjinv00 = jhjinv[d,t,f,aa,0,0]
                    jhjinv01 = jhjinv[d,t,f,aa,0,1]
                    jhjinv10 = jhjinv[d,t,f,aa,1,0]
                    jhjinv11 = jhjinv[d,t,f,aa,1,1]
                    
                    upd[d,t,f,aa,0,0] = (jhr00*jhjinv00 + jhr01*jhjinv10) 
                    upd[d,t,f,aa,0,1] = (jhr00*jhjinv01 + jhr01*jhjinv11) 
                    upd[d,t,f,aa,1,0] = (jhr10*jhjinv00 + jhr11*jhjinv10)
                    upd[d,t,f,aa,1,1] = (jhr10*jhjinv01 + jhr11*jhjinv11)

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_corrected(o, g, gh, corr, t_int, f_int):
    """
    Given the observed visbilities, inverse gains, and their conjugates, computes the corrected
    visibilitites.

    Args:
        o (np.complex64 or np.complex128):
            Typed memoryview of observed visibility array with dimensions (t, f, a, a, c, c).
        g (np.complex64 or np.complex128):
            Typed memoryview of inverse gain array with dimensions (d, ti, fi, a, c, c).
        gh (np.complex64 or np.complex128):
            Typed memoryview of inverse conjugate gain array with dimensions (d, ti, fi, a, c, c).
        corr (np.complex64 or np.complex128):
            Typed memoryview of corrected data array with dimensions (t, f, a, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = g.shape[0]
    n_tim = o.shape[0]
    n_fre = o.shape[1]
    n_ant = o.shape[2]

    bls = np.array([[i,j] for i in range(n_ant) for j in range(i, n_ant)], dtype=np.int32)
    n_bl = bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])

    for ibl in prange(n_bl):
        aa, ab = bls[ibl][0], bls[ibl][1]
        for t in range(n_tim):
            bt = broadcast_times[t]
            for f in range(n_fre):
                bf = broadcast_freqs[f]

                g00 = g[0,bt,bf,aa,0,0]
                g01 = g[0,bt,bf,aa,0,1]
                g10 = g[0,bt,bf,aa,1,0]
                g11 = g[0,bt,bf,aa,1,1]

                o00 = o[t,f,aa,ab,0,0]
                o01 = o[t,f,aa,ab,0,1]
                o10 = o[t,f,aa,ab,1,0]
                o11 = o[t,f,aa,ab,1,1]

                gh00 = gh[0,bt,bf,ab,0,0]
                gh01 = gh[0,bt,bf,ab,0,1]
                gh10 = gh[0,bt,bf,ab,1,0]
                gh11 = gh[0,bt,bf,ab,1,1]

                corr[t,f,aa,ab,0,0] = ((g00*o00 + g01*o10)*gh00 + (g00*o01 + g01*o11)*gh10)
                corr[t,f,aa,ab,0,1] = ((g00*o00 + g01*o10)*gh01 + (g00*o01 + g01*o11)*gh11)
                corr[t,f,aa,ab,1,0] = ((g10*o00 + g11*o10)*gh00 + (g10*o01 + g11*o11)*gh10)
                corr[t,f,aa,ab,1,1] = ((g10*o00 + g11*o10)*gh01 + (g10*o01 + g11*o11)*gh11)

                corr[t,f,ab,aa,0,0] = corr[t,f,aa,ab,0,0].conjugate()
                corr[t,f,ab,aa,1,0] = corr[t,f,aa,ab,0,1].conjugate()
                corr[t,f,ab,aa,0,1] = corr[t,f,aa,ab,1,0].conjugate()
                corr[t,f,ab,aa,1,1] = corr[t,f,aa,ab,1,1].conjugate()

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def apply_gains(m, g, gh, t_int, f_int):
    """
    Applies the gains and their conjugates to the model array. This operation is performed in place
    - be wary of losing the original array. The result has full time and frequency resolution -
    solution intervals are used to correctly associate the gains with the model.

    Args:
        m (np.complex64 or np.complex128):
            Typed memoryview of model visibility array with dimensions (d, m , t, f, a, a, c, c).
        g (np.complex64 or np.complex128):
            Typed memoryview of gain array with dimensions (d, ti, fi, a, c, c).
        gh (np.complex64 or np.complex128):
            Typed memoryview of conjugate gain array with dimensions (d, ti, fi, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = m.shape[0]
    n_mod = m.shape[1]
    n_tim = m.shape[2]
    n_fre = m.shape[3]
    n_ant = m.shape[4]

    g_dir = g.shape[0]

    bls = np.array([[i,j] for i in range(n_ant) for j in range(i+1, n_ant)], dtype=np.int32)
    n_bl = bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])
    broadcast_dirs = np.array([d%g_dir for d in range(n_dir)])

    for ibl in prange(n_bl):
        aa, ab = bls[ibl][0], bls[ibl][1]  
        for i in range(n_mod):
            for t in range(n_tim):
                bt = broadcast_times[t]
                for f in range(n_fre):
                    bf = broadcast_freqs[f]
                    for d in range(n_dir):
                        bd = broadcast_dirs[d]
                        
                        g00 = g[bd,bt,bf,aa,0,0]
                        g01 = g[bd,bt,bf,aa,0,1]
                        g10 = g[bd,bt,bf,aa,1,0]
                        g11 = g[bd,bt,bf,aa,1,1]

                        m00 = m[d,i,t,f,aa,ab,0,0]
                        m01 = m[d,i,t,f,aa,ab,0,1]
                        m10 = m[d,i,t,f,aa,ab,1,0]
                        m11 = m[d,i,t,f,aa,ab,1,1]

                        gh00 = gh[bd,bt,bf,ab,0,0]
                        gh01 = gh[bd,bt,bf,ab,0,1]
                        gh10 = gh[bd,bt,bf,ab,1,0]
                        gh11 = gh[bd,bt,bf,ab,1,1]

                        tmp00 = ((g00*m00 + g01*m10)*gh00 + (g00*m01 + g01*m11)*gh10)
                        tmp01 = ((g00*m00 + g01*m10)*gh01 + (g00*m01 + g01*m11)*gh11)
                        tmp10 = ((g10*m00 + g11*m10)*gh00 + (g10*m01 + g11*m11)*gh10)
                        tmp11 = ((g10*m00 + g11*m10)*gh01 + (g10*m01 + g11*m11)*gh11)

                        m[d,i,t,f,aa,ab,0,0] = tmp00
                        m[d,i,t,f,aa,ab,0,1] = tmp01
                        m[d,i,t,f,aa,ab,1,0] = tmp10
                        m[d,i,t,f,aa,ab,1,1] = tmp11

                        m[d,i,t,f,ab,aa,0,0] = tmp00.conjugate()
                        m[d,i,t,f,ab,aa,1,0] = tmp01.conjugate()
                        m[d,i,t,f,ab,aa,0,1] = tmp10.conjugate()
                        m[d,i,t,f,ab,aa,1,1] = tmp11.conjugate()

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def right_multiply_gains(g, g_next, t_int, f_int):
    """
    Multiples two gain terms in place. Result has full time and frequency resolution 
    even if g_next does does not. For use in Jones chain.

    NOTE: THIS MAY BE INCORRECT/BACKWARDS.

    Args:
        g (np.complex64 or np.complex128):
            Typed memoryview of gain array with dimensions (d, t, f, a, c, c).
        g_next (np.complex64 or np.complex128):
            Typed memoryview of gain array with dimensions (d, ti, fi, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """

    n_dir = g.shape[0]
    n_tim = g.shape[1]
    n_fre = g.shape[2]
    n_ant = g.shape[3]

    g_dir = g_next.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])
    broadcast_dirs = np.array([d%g_dir for d in range(n_dir)])

    for aa in prange(n_ant):
        for t in range(n_tim):
            bt = broadcast_times[t]
            for f in range(n_fre):
                bf = broadcast_freqs[f]
                for d in range(n_dir):
                    bd = broadcast_dirs[d]

                    g00 = g[d,t,f,aa,0,0]
                    g01 = g[d,t,f,aa,0,1]
                    g10 = g[d,t,f,aa,1,0]
                    g11 = g[d,t,f,aa,1,1]

                    g_next00 = g_next[bd,bt,bf,aa,0,0]
                    g_next01 = g_next[bd,bt,bf,aa,0,1]
                    g_next10 = g_next[bd,bt,bf,aa,1,0]
                    g_next11 = g_next[bd,bt,bf,aa,1,1]

                    g[d,t,f,aa,0,0] = (g00*g_next00 + g01*g_next10)
                    g[d,t,f,aa,0,1] = (g00*g_next01 + g01*g_next11)
                    g[d,t,f,aa,1,0] = (g10*g_next00 + g11*g_next10)
                    g[d,t,f,aa,1,1] = (g10*g_next01 + g11*g_next11)

# Map the J^H.J inversion method to a generic inversion.
compute_jhjinv = generics.compute_2x2_inverse

# Map inversion to generic 2x2 inverse.
invert_gains = generics.compute_2x2_inverse



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details

# This module has been adapted from the DDFacet package,
# (c) Cyril Tasse et al., see http://github.com/saopicc/DDFacet

from __future__ import print_function
from six import string_types
import logging, logging.handlers, os, re, sys, multiprocessing, time, signal
import psutil
from . import ModColor

# dict of logger wrappers created by the application
_loggers = {}

# global verbosity levels (used for loggers for which an explicit level is not set)
_global_verbosity = 0
_global_log_verbosity = None

# this will be the handler for the log file
_file_handler = None
# this will be a null handler
_null_handler = logging.NullHandler()

def logToFile(filename, append=False):
    global _file_handler
    if not _file_handler:
        _file_handler = logging.FileHandler(filename, mode='a' if append else 'w')
        _file_handler.setLevel(logging.DEBUG)
        _file_handler.setFormatter(_logfile_formatter)
        # set it as the target for the existing wrappers' handlers
        for wrapper in _loggers.values():
            wrapper.logfile_handler.setTarget(_file_handler)

def getLogFilename():
    '''Returns log filename if logToFile has been called previously, None otherwise'''
    global _file_handler
    if not _file_handler:
        return None
    return _file_handler.baseFilename

class _DefaultWriter(object):
    """A default writer logs messages to a logger"""
    __print_once_keys = []
    def __init__(self, logger, level, color=None, bold=None):
        self.logger = logger
        self.level = level
        self.color = (color or "red") if bold else color
        self.bold = bool(color) if bold is None else bold

    def write(self, message, level_override=None, print_once=None):
        if print_once is not None:
            if print_once in _DefaultWriter.__print_once_keys:
                return
            _DefaultWriter.__print_once_keys = list(set(_DefaultWriter.__print_once_keys + [print_once]))

        message = message.rstrip()
        if self.color and message:  # do not colorize empty messages, else "\n" is issued independently
            message = ModColor.Str(message, col=self.color, Bold=self.bold)
        self.logger.log(self.level if level_override is None else level_override, message)

    def print(self, *args):
        return self.write(" ".join(map(str, args)))

class LoggerWrapper(object):
    def __init__(self, logger, verbose=None, log_verbose=None):
        self.logger = logger
        logger.propagate = False

        # initialize handlers for console and logfile

        self.console_handler = logging.StreamHandler(sys.stderr)
        self.console_handler.setFormatter(_console_formatter)

        self.logfile_handler = logging.handlers.MemoryHandler(1, logging.DEBUG, _file_handler or _null_handler)
        self.logfile_handler.setFormatter(_logfile_formatter)

        # set verbosity levels
        self._verbose = self._log_verbose = None
        self.verbosity(verbose if verbose is not None else _global_verbosity)
        self.log_verbosity(log_verbose if log_verbose is not None else _global_log_verbosity)

        # other init

        self.logger.addHandler(self.console_handler)
        self.logger.addHandler(self.logfile_handler)
        self.logger.addFilter(_log_filter)

    def verbosity(self, set_verb=None):
        if set_verb is not None:
            self._verbose = set_verb
            self.console_handler.setLevel(logging.INFO - set_verb)
            if self._log_verbose is None:
                self.logfile_handler.setLevel(logging.INFO - set_verb)
        return self._verbose

    def log_verbosity(self, set_verb=None):
        if set_verb is not None:
            self._log_verbose = set_verb
            self.logfile_handler.setLevel(logging.INFO - set_verb)
        return self._log_verbose if self._log_verbose is not None else self._verbose

    def __call__(self, level, color=None):
        """
        Function call operator on logger. Use to issue messages at different verbosity levels.
        E.g. log(2).print("message" will issue a message at level logging.INFO - 2.)
        An optional color argument will colorize the message.

        Returns:
            A writer object (to which a message may be sent with "<<")
        """
        # effective verbosity level is either set explicitly when the writer is created, or else use global level
        return _DefaultWriter(self.logger, logging.INFO - level, color=color)
    
    def warn(self, msg, color=None, print_once=None):
        """
        Wrapper for log.warn
        """
        _DefaultWriter(self.logger, logging.WARN, color=color).write(msg, print_once=print_once)

    warning = warn

    def error(self, msg, color="red", print_once=None):
        """
        Wrapper for log.error
        """
        _DefaultWriter(self.logger, logging.ERROR, color=color).write(msg, print_once=print_once)
        
    def info(self, msg, color=None, print_once=None):
        """
        Wrapper for log.info
        """
        _DefaultWriter(self.logger, logging.INFO, color=color).write(msg, print_once=print_once)

    def critical(self, msg, color=None, print_once=None):
        """
        Wrapper for log.critical
        """
        _DefaultWriter(self.logger, logging.CRITICAL, color=color).write(msg, print_once=print_once)

    def debug(self, msg, color=None, print_once=None):
        """
        Wrapper for log.debug
        """
        _DefaultWriter(self.logger, logging.DEBUG, color=color).write(msg, print_once=print_once)

    def exception(self, msg, color=None, print_once=None):
        """
        Wrapper for log.exception
        """
        _DefaultWriter(self.logger, logging.EXCEPTION, color=color).write(msg, print_once=print_once)

    def print(self, *args):
        return self.info(" ".join(map(str, args)))

    def write(self, message, level=logging.INFO, verbosity=0, print_once=None, color=None):
        # apply verbosity only to INFO levels
        if level == logging.INFO:
            level -= int(verbosity)
        _DefaultWriter(self.logger, level, color=color).write(message,
                                                                               print_once=print_once)

_parent_process = psutil.Process(os.getpid())
_log_memory_totals = True
_log_memory_types = None, "rss vms".split(), "rss pss".split()
GB = float(1024 ** 3)


def enableMemoryLogging(level=1):
    LogFilter.setMemoryLogging((level or 0)%3)   # level is 0/1/2

_subprocess_label = None

def set_subprocess_label(label):
    """
    Sets the subprocess label explicitly
    """
    global _subprocess_label
    _subprocess_label = label


def get_subprocess_label():
    """
    Returns the subprocess ID. For the main process, this is empty. For subprocesses
    (started by multiprocessing), this is "Pn" by default, where n is the process number.
    """
    global _subprocess_label
    if _subprocess_label is None:
        name = multiprocessing.current_process().name
        if name == "MainProcess":
            _subprocess_label = ""
        else:
            _subprocess_label = name.replace("Process-", "P")
    return _subprocess_label


def _sigusr1_handler(signum, frame):
    level = 2 if LogFilter._log_memory == 1 else 1
    print("pid {} received USR1: memory logging level {}".format(os.getpid(), level))
    LogFilter.setMemoryLogging(level)


def _sigusr2_handler(signum, frame):
    print("pid {} received USR2: disabling memory logging".format(os.getpid()))
    LogFilter.setMemoryLogging(0)


signal.signal(signal.SIGUSR1, _sigusr1_handler)
signal.signal(signal.SIGUSR2, _sigusr2_handler)

class LogFilter(logging.Filter):
    _log_memory = 0


    @staticmethod
    def setMemoryLogging(level):
        LogFilter._log_memory = level
        LogFilter._mem_ts = LogFilter._mem_totals_ts = LogFilter._children_ts = 0

    _mem = None
    _mem_totals = None
    _children = None
    _mem_ts = 0
    _mem_totals_ts = 0
    _children_ts = 0
    _mem_update = (0, .5, 2)     # full memory updates are a little costly, so do them only after this many seconds has elapsed
    _mem_totals_update = (0, 1, 5)
    _children_update = 3 # children updates even more so, so do it even less frequently

    """LogFilter augments the event by a few new attributes used by our formatter"""
    def filter(self, event):
        if not event.getMessage().strip():
            return False
        # short logger name (without app_name in front of it)
        setattr(event, 'shortname', event.name.split('.',1)[1] if '.' in event.name else event.name)
        setattr(event, 'separator', '| ')
        memlevel = LogFilter._log_memory  # signal handler can change that midway through, so use this value
        if memlevel:
            # memory usage info
            t = time.time()
            if t - self._mem_ts > self._mem_update[memlevel]:
                KEYS = _log_memory_types[memlevel]
                # get memory for this process
                mi0 = psutil.Process(os.getpid()).memory_full_info() if memlevel == 2 else \
                    psutil.Process(os.getpid()).memory_info()
                mem = {key: getattr(mi0, key) / GB for key in KEYS}
                shm = psutil.virtual_memory().shared / GB
                # get total memory
                if _log_memory_totals:
                    if t - self._mem_totals_ts > self._mem_totals_update[memlevel]:
                        if t - self._children_ts > self._children_update:
                            self._children = [_parent_process] + list(_parent_process.children(recursive=True))
                            self._children_ts = time.time()
                        if len(self._children) > 1:
                            mis = []
                            # scan over children, ignoring ones that may have disappeared
                            for p in self._children:
                                try:
                                    mis.append(p.memory_full_info()) if memlevel == 2 else mis.append(p.memory_info())
                                except psutil.NoSuchProcess:
                                    pass
                            self._mem_totals = {key: sum([getattr(mi,key) for mi in mis]) / GB for key in KEYS}
                        self._mem_totals_ts = time.time()
                # form up string
                if self._mem_totals is None:
                    smem = ["{:.1f}".format(mem.get(key, 0)) for key in KEYS]
                else:
                    smem = ["{:.1f}/{:.1f}".format(mem.get(key, 0), self._mem_totals.get(key, 0)) for key in KEYS]
                smem.append("{:.1f}Gb".format(shm))
                self._mem = " ".join(smem)
                self._mem_ts = time.time()
            setattr(event, "memory", "[{}{}] ".format("*" if memlevel == 2 else "", self._mem))
            setattr(event, 'separator', '')
        else:
            setattr(event, "memory", "")
        # subprocess info
        subprocess_id = get_subprocess_label()
        if subprocess_id:
            setattr(event, "subprocess", ModColor.Str("[%s] "%subprocess_id, col="blue"))
            setattr(event, 'separator', '')
        else:
            setattr(event, "subprocess", "")
        return True

_log_filter = LogFilter()

class ColorStrippingFormatter(logging.Formatter):
    def __init__(self, fmt, datefmt, strip=True):
        logging.Formatter.__init__(self, fmt, datefmt)
        self.strip = strip

    def label(self, record):
        if record.levelno < logging.WARNING:
            return "INFO", "37;42"
        elif record.levelno < logging.ERROR:
            return "WARNING", "37;43"
        elif record.levelno < logging.CRITICAL:
            return "ERROR", "5;41"
        else:
            return "CRITICAL", "5;41"


    def format(self, record):
        """Uses parent class to format record, then strips off colors"""
        msg = logging.Formatter.format(self, record)
        label, color = self.label(record)
        if self.strip:
            return "{:10s}{}".format(label, re.sub("\033\\[[0-9]+m", "", msg, 0))
        else:
            return "\033[1;{}m{:10s}\033[0m{}".format(color, label, msg)\

_fmt = "%(asctime)s - %(shortname)-18.18s %(subprocess)s%(memory)s%(separator)s%(message)s"
#        _fmt = "%(asctime)s %(name)-25.25s | %(message)s"
_datefmt = '%H:%M:%S'#'%H:%M:%S.%f'
_logfile_formatter = ColorStrippingFormatter(_fmt, _datefmt, strip=True)
_console_formatter = ColorStrippingFormatter(_fmt, _datefmt, strip=False)

_root_logger = None
log = None

def init(app_name):
    global _app_name
    global _root_logger
    if _root_logger is None:
        # logging.basicConfig(level=logging.DEBUG, datefmt=_datefmt)
        logging.basicConfig(datefmt=_datefmt)
        _app_name = app_name
        _root_logger = logging.getLogger(app_name)
        _root_logger.setLevel(logging.DEBUG)
        global log
        log = _loggers[''] = LoggerWrapper(_root_logger)

def getLogger(name, verbose=None, log_verbose=None):
    """Creates a new logger (or returns one, if already created)"""
    init("app")
    if name in _loggers:
        return _loggers[name]

    logger = logging.getLogger("{}.{}".format(_app_name, name))
    lw = _loggers[name] = LoggerWrapper(logger, verbose, log_verbose)
    lw(2).print("logger initialized")

    return lw

def setBoring(boring=True):
    _console_formatter.strip = boring

def setGlobalVerbosity(verbosity):
    global _global_verbosity
    if verbosity is None:
        _global_verbosity = 0
        return
    # ensure verbosity is turned into a list.
    if type(verbosity) is int:
        verbosity = [verbosity]
    elif isinstance(verbosity, string_types):
        verbosity = verbosity.split(",")
    elif not isinstance(verbosity, (list, tuple)):
        raise TypeError("can't parse verbosity specification of type '{}'".format(type(verbosity)))
    for element in verbosity:
        if type(element) is int or re.match("^[0-9]+$", element):
            _global_verbosity = int(element)
            log(0, "green").print("set global console verbosity level {}".format(_global_verbosity))
        else:
            m = re.match("^(.+)=([0-9]+)$", element)
            if not m:
                raise ValueError("can't parse verbosity specification '{}'".format(element))
            logger = getLogger(m.group(1))
            level = int(m.group(2))
            logger.verbosity(level)
            logger(0,"green").print("set console verbosity level {}={}".format(m.group(1), level))

def setGlobalLogVerbosity(verbosity):
    global _global_log_verbosity
    if verbosity is None:
        _global_log_verbosity = None  # None means follow console default
        return
    # ensure verbosity is turned into a list.
    if type(verbosity) is int:
        verbosity = [verbosity]
    elif isinstance(verbosity, string_types):
        verbosity = verbosity.split(",")
    elif not isinstance(verbosity, (list, tuple)):
        raise TypeError("can't parse verbosity specification of type '{}'".format(type(verbosity)))
    for element in verbosity:
        if type(element) is int or re.match("^[0-9]+$", element):
            _global_log_verbosity = int(element)
            if _global_log_verbosity is not None:
                log(0, "green").print("set global log verbosity level {}".format(_global_log_verbosity))
        else:
            m = re.match("^(.+)=([0-9]+)$", element)
            if not m:
                raise ValueError("can't parse verbosity specification '{}'".format(element))
            logger = getLogger(m.group(1))
            level = int(m.group(2))
            logger.log_verbosity(level)
            logger(0,"green").print("set log verbosity level {}={}".format(m.group(1), level))


def setSilent(Lname):
    """Silences the specified sublogger(s)"""
    log.print(ModColor.Str("set silent: %s" % Lname, col="red"))
    if isinstance(Lname, string_types):
        getLogger(Lname).logger.setLevel(logging.CRITICAL)
    elif type(Lname) is list:
        for name in Lname:
            getLogger(name).logger.setLevel(logging.CRITICAL)


def setLoud(Lname):
    """Un-silences the specified sublogger(s)"""
    log.print(ModColor.Str("set loud: %s" % Lname, col="green"))
    if isinstance(Lname, string_types):
        getLogger(Lname).logger.setLevel(logging.DEBUG)
    elif type(Lname) is list:
        for name in Lname:
            getLogger(name).logger.setLevel(logging.DEBUG)


if __name__=="__main__":
    log=getLogger("a.x")
    log.print("a.x")



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017 SKA South Africa
#
# This file is part of CubeCal.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import cubical
from setuptools import setup, find_packages

with open('README.md') as f:
    long_description = f.read()

try:
    import six
except ImportError:
    raise ImportError("Please install six before running install. If you're using pip 19 to install this package you should not be seeing this message.")

try:
    import numpy
except ImportError:
    raise ImportError("Please install numpy before running install. If you're using pip 19 to install this package you should not be seeing this message.")

# Check for readthedocs environment variable.

on_rtd = os.environ.get('READTHEDOCS') == 'True'

if on_rtd:
    requirements = ['numpy',
                    'matplotlib', 
                    'scipy']
else:
    requirements = ['future',
                    'numpy',
                    'numba',
                    'python-casacore',
                    'sharedarray >= 3.2.1', 
                    'matplotlib',
                    'scipy',
                    'astro-tigger-lsm',
                    'six',
                    'astropy>=3.0',
                    'psutil'
                    ]

setup(name='cubical',
      version=cubical.VERSION,
      description='Fast calibration implementation exploiting complex optimisation.',
      url='https://github.com/ratt-ru/CubiCal',
      classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Astronomy"],
      author='Jonathan Kenyon',
      author_email='jonosken@gmail.com',
      license='GNU GPL v3',
      long_description=long_description,
      long_description_content_type='text/markdown',
      packages=find_packages(),
      python_requires=">=3.6", 
      install_requires=requirements,
      include_package_data=True,
      zip_safe=False,
      scripts=['cubical/bin/print-cubical-stats',
               'cubical/bin/plot-leakage-solutions',
               'cubical/bin/plot-gain-solutions'],
      entry_points={'console_scripts': ['gocubical = cubical.main:main']},
      extras_require={
          'lsm-support': ['montblanc >= 0.6.4'],
          'degridder-support': ['ddfacet >= 0.6.0', 
                                'regions >= 0.4',
                                'meqtrees-cattery >= 1.7.0']
      }
)



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
from __future__ import print_function
from builtins import range
from cubical.machines.abstract_machine import MasterMachine
from cubical.machines.complex_2x2_machine import Complex2x2Gains
from cubical.machines.complex_W_2x2_machine import ComplexW2x2Gains
import numpy as np
import cubical.kernels

from cubical.tools import logger
from . import machine_types
from cubical.flagging import FL
log = logger.getLogger("jones_chain")

class JonesChain(MasterMachine):
    """
    This class implements a gain machine for an arbitrary chain of Jones matrices. Most of its
    functionality is consistent with a complex 2x2 solver - many of its methods mimic those of the 
    underlying complex 2x2 machines.
    """

    def __init__(self, label, data_arr, ndir, nmod, times, frequencies, chunk_label, jones_options):
        """
        Initialises a chain of complex 2x2 gain machines.
        
        Args:
            label (str):
                Label identifying the Jones term.
            data_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed 
                visibilities. 
            ndir (int):
                Number of directions.
            nmod (nmod):
                Number of models.
            times (np.ndarray):
                Times for the data being processed.
            frequencies (np.ndarray):
                Frequencies for the data being processsed.
            jones_options (dict): 
                Dictionary of options pertaining to the chain. 
        """
        from cubical.main import UserInputError
        # This instantiates the number of complex 2x2 elements in our chain. Each element is a
        # gain machine in its own right - the purpose of this machine is to manage these machines
        # and do the relevant fiddling between parameter updates. When combining DD terms with
        # DI terms, we need to be initialise the DI terms using only one direction - we do this with
        # slicing rather than summation as it is slightly faster.
        self.jones_terms = []
        self.num_left_di_terms = 0  # how many DI terms are there at the left of the chain
        seen_dd_term = False

        for iterm, term_opts in enumerate(jones_options['chain']):
            jones_class = machine_types.get_machine_class(term_opts['type'])
            if jones_class is None:
                raise UserInputError("unknown Jones class '{}'".format(term_opts['type']))
            if not issubclass(jones_class, Complex2x2Gains) and not issubclass(jones_class, ComplexW2x2Gains) and term_opts['solvable']:
                raise UserInputError("only complex-2x2 or robust-2x2 terms can be made solvable in a Jones chain")
            term = jones_class(term_opts["label"], data_arr, ndir, nmod, times, frequencies, chunk_label, term_opts)
            self.jones_terms.append(term)
            if term.dd_term:
                seen_dd_term = True
            elif not seen_dd_term:
                self.num_left_di_terms = iterm


        MasterMachine.__init__(self, label, data_arr, ndir, nmod, times, frequencies,
                               chunk_label, jones_options)

        self.chain  = cubical.kernels.import_kernel("chain")
        # kernel used for compute_residuals and such
        self.kernel = Complex2x2Gains.get_full_kernel(jones_options, diag_gains=self.is_diagonal)

        self.n_dir, self.n_mod = ndir, nmod
        _, self.n_tim, self.n_fre, self.n_ant, self.n_ant, self.n_cor, self.n_cor = data_arr.shape


        self.n_terms = len(self.jones_terms)
        # make list of number of iterations per solvable term
        # If not specified, just use the maxiter setting of each term
        # note that this list is updated as we converge, so make a copy
        term_iters = jones_options['sol']['term-iters']
        if not term_iters:
            self.term_iters = [term.maxiter for term in self.jones_terms if term.solvable]
        elif type(term_iters) is int:
            self.term_iters = [term_iters]
        elif isinstance(term_iters, (list, tuple)):
            self.term_iters = list(term_iters)
        else:
            raise UserInputError("invalid term-iters={} setting".format(term_iters))

        self.solvable = bool(self.term_iters) and any([term.solvable for term in self.jones_terms])

        # setup first solvable term in chain
        self.active_index = None

        # this list accumulates the per-term convergence status strings
        self._convergence_states = []
        # True when the last active term has had its convergence status queried
        self._convergence_states_finalized = False

        self.cached_model_arr = self._r = self._m = None

    @classmethod
    def determine_diagonality(cls, options):
        """Returns true if the machine class, given the options, represents a diagonal gain"""
        from cubical.main import UserInputError
        diagonal = True
        for term_opts in options['chain']:
            jones_class = machine_types.get_machine_class(term_opts['type'])
            if jones_class is None:
                raise UserInputError("unknown Jones class '{}'".format(term_opts['type']))
            diagonal = diagonal and jones_class.determine_diagonality(term_opts)
        return diagonal

    @classmethod
    def determine_allocators(cls, options):
        kernel = Complex2x2Gains.get_full_kernel(options, diag_gains=cls.determine_diagonality(options))
        return kernel.allocate_vis_array, kernel.allocate_flag_array, kernel.allocate_gain_array

    def precompute_attributes(self, data_arr, model_arr, flags_arr, inv_var_chan):
        """Precomputes various stats before starting a solution"""
        MasterMachine.precompute_attributes(self, data_arr, model_arr, flags_arr, inv_var_chan)
        for term in self.jones_terms:
            term.precompute_attributes(data_arr, model_arr, flags_arr, inv_var_chan)

    def export_solutions(self):
        """ Saves the solutions to a dict of {label: solutions,grids} items. """

        soldict = {}
        # prefix jones label to solution name
        for term in self.jones_terms:
            if term.solvable:
                for label, sol in term.export_solutions().items():
                    soldict["{}:{}".format(term.jones_label, label)] = sol
        soldict['prefixed'] = True

        return soldict

    def importable_solutions(self, grid):
        """ Returns a dictionary of importable solutions for the chain. """

        soldict = {}
        for term in self.jones_terms:
            soldict.update(term.importable_solutions(grid))

        return soldict

    def import_solutions(self, soldict):
        """
        Loads solutions from a dict. This should not be called -- _load_solutions()
        below should rather call import_solutions() on all the chain terms.
        """
        raise RuntimeError("This method cannot be called on a Jones chain. This is a bug.")

    def _load_solutions(self, init_sols, full_grid):
        """
        Helper method invoked by Factory.create_machine() to import existing solutions into machine.
        
        In the case of a chain, we invoke this method on every member.
        """
        for term in self.jones_terms:
            term._load_solutions(init_sols, full_grid)

    #@profile
    def compute_js(self, obser_arr, model_arr):
        """
        This function computes the (J\ :sup:`H`\J)\ :sup:`-1` and J\ :sup:`H`\R terms of the GN/LM 
        method. This method is more complicated than a more conventional gain machine. The use of
        a chain means there are additional terms which need to be considered when computing the 
        parameter updates.

        Args:
            obser_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                observed visibilities.
            model_arr (np.ndrray): 
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                model visibilities.

        Returns:
            3-element tuple
                
                - J\ :sup:`H`\R (np.ndarray)
                - (J\ :sup:`H`\J)\ :sup:`-1` (np.ndarray)
                - Count of flags raised (int)     
        """     

        n_dir, n_tint, n_fint, n_ant, n_cor, n_cor = self.active_term.gains.shape

        if self.last_active_index!=self.active_index or self.iters==1:
            self.cached_model_arr = cached_model_arr = np.empty_like(model_arr)
            np.copyto(cached_model_arr, model_arr)

            for ind in range(self.n_terms - 1, self.active_index, -1):
                term = self.jones_terms[ind]
                term.apply_gains(cached_model_arr, full2x2=True)

            # collapse direction axis, if current term is non-DD
            if not self.active_term.dd_term and self.n_dir>1:
                self.cached_model_arr = np.empty_like(model_arr[0:1,...])
                np.sum(cached_model_arr, axis=0, keepdims=True, out=self.cached_model_arr)

            self.jh = np.empty_like(self.cached_model_arr)

            jhr_shape = [n_dir if self.active_term.dd_term else 1, self.n_tim, self.n_fre, n_ant, n_cor, n_cor]

            self._jhr = self.active_term.allocate_gain_array(jhr_shape, dtype=obser_arr.dtype)

            jhrint_shape = [n_dir, n_tint, n_fint, n_ant, n_cor, n_cor]

            self._jhrint = self.active_term.allocate_gain_array(jhrint_shape, dtype=self._jhr.dtype)
            self._jhj = np.empty_like(self._jhrint)
            self._jhjinv =  np.empty_like(self._jhrint)

        np.copyto(self.jh, self.cached_model_arr)

        for ind in range(self.active_index, -1, -1):
            term = self.jones_terms[ind]
            self.chain.compute_jh(self.jh, term.gains, *term.gain_intervals)
            
        if n_dir > 1:
            if self._r is None:
                self._r = np.empty_like(obser_arr)
            r = self.compute_residual(obser_arr, model_arr, self._r, require_full=False)
        else:
            r = obser_arr

        if self.active_term.offdiag_only:
            # self.jh[..., (0, 1), (0, 1)] = 0
            if r is obser_arr:
                r = r.copy()
            r[..., (0, 1), (0, 1)] = 0

        if self.active_term.diag_only:
            # self.jh[..., (0, 1), (1, 0)] = 0
            if r is obser_arr:
                r = r.copy()
            r[..., (0, 1), (1, 0)] = 0

        self._jhr.fill(0)

        #if self.active_term.jones_label == "D":
        #    import ipdb; ipdb.set_trace()

        self.active_term.kernel.compute_jhr(self.jh, r, self._jhr, 1, 1)


        for ind in range(0, self.active_index, 1):
            term = self.jones_terms[ind]
            g_inv, gh_inv, flag_counts = term.get_inverse_gains()
            self.chain.apply_left_inv_jones(self._jhr, g_inv, *term.gain_intervals)

        self._jhrint.fill(0)
        self.chain.sum_jhr_intervals(self._jhr, self._jhrint, *self.active_term.gain_intervals)

        self._jhj.fill(0)
        self.active_term.kernel.compute_jhj(self.jh, self._jhj, *self.active_term.gain_intervals)

        flag_count = self.active_term.kernel.compute_jhjinv(self._jhj, self._jhjinv,
                                                    self.active_term.gflags, self.active_term.eps, FL.ILLCOND)

        #if self.active_term.jones_label == "D":
        #    import ipdb; ipdb.set_trace()

        return self._jhrint, self._jhjinv, flag_count

    def implement_update(self, jhr, jhjinv):
        return self.active_term.implement_update(jhr, jhjinv)

    def accumulate_gains(self, dd=True):
        """
        This function returns the product of all the gains in the chain, at full TF resolution,
        for all directions (dd=True), or just the 0th direction.

        Args:
            dd (bool):
                Accumulate per-direction gains, if available. If False, only the
                first direction is used.

        Returns:
            A tuple of gains,conjugate gains
        """
        ndir = self.n_dir if dd else 1
        gains = self.jones_terms[0].allocate_gain_array([ndir, self.n_tim, self.n_fre, self.n_ant, self.n_cor, self.n_cor],
                                                        self.dtype)
        g0 = self.jones_terms[0]._gainres_to_fullres(self.jones_terms[0].gains, tdim_ind=1)
        if ndir > 1 and g0.shape[0] == 1:
            g0 = g0.reshape(g0.shape[1:])[np.newaxis,...]
        elif ndir == 1 and g0.shape[0] > 1:
            g0 = g0[:1,...]
        gains[:] = g0
        for term in self.jones_terms[1:]:
            term.kernel.right_multiply_gains(gains, term.gains, *term.gain_intervals)

        # compute conjugate gains
        gh = np.empty_like(gains)
        np.conj(gains.transpose(0, 1, 2, 3, 5, 4), gh)

        return gains, gh

    def accumulate_inv_gains(self, direction=None):
        """
        This function returns the inverse of the product of all the gains in the chain, at full TF resolution.
        If direction=None, only DI gains are included. Otherwise, also includes DD gains for the given direction.

        Returns:
            A tuple of gains,conjugate gains,flag_count (if flags raised in inversion)
        """
        gains = self.jones_terms[-1].allocate_gain_array([1, self.n_tim, self.n_fre, self.n_ant, self.n_cor, self.n_cor],
                                                  self.dtype)
        init = False
        fc0 = 0
        # flip order of jones terms for inverse
        for term in self.jones_terms[::-1]:
            if direction is not None or not term.dd_term:
                # dirslice will select the specified direction from the GF array
                dirslice = slice(0, 1) if not term.dd_term else slice(direction, direction + 1)
                g, _, fc = term.get_inverse_gains()
                # g = term._gainres_to_fullres(g, tdim_ind=1)
                fc0 += fc
                if init:
                    term.kernel.right_multiply_gains(gains, g[dirslice,...], *term.gain_intervals)
                else:
                    init = True
                    gains[:] = term._gainres_to_fullres(g[dirslice,...], tdim_ind=1)

        # compute conjugate gains
        gh = np.empty_like(gains)
        np.conj(gains.transpose(0, 1, 2, 3, 5, 4), gh)

        return gains, gh, fc


    #@profile
    def compute_residual(self, obser_arr, model_arr, resid_arr, require_full=True):
        """
        This function computes the residual. This is the difference between the
        observed data, and the model data with the gains applied to it.

        Args:
            obser_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                observed visibilities.
            model_arr (np.ndrray): 
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                model visibilities.
            resid_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array into which the 
                computed residuals should be placed.
            require_full (bool):
                If True, a full 2x2 residual is required. If False, only the terms used in the solution
                (e.g. if a solution is diagonal, only the diagonals) are required.

        Returns:
            np.ndarray: 
                Array containing the result of computing D - GMG\ :sup:`H`.
        """
        g, gh = self.accumulate_gains()
        np.copyto(resid_arr, obser_arr)
        kernel = self.kernel if require_full else self.active_term.kernel_solve
        kernel.compute_residual(model_arr, g, gh, resid_arr, 1, 1)

        return resid_arr

    def apply_inv_gains(self, resid_vis, corr_vis=None, full2x2=True, direction=None):
        """
        Applies the inverse of the gain estimates to the observed data matrix.

        Args:
            obser_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                observed visibilities.
            corr_vis (np.ndarray or None, optional): 
                if specified, shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array 
                into which the corrected visibilities should be placed.
            direction (int or None):
                if None, only DI terms will be applied. Otherwise, also applies DD terms for given direction.

        Returns:
            np.ndarray: 
                Array containing the result of G\ :sup:`-1`\DG\ :sup:`-H`.
        """

        if corr_vis is None:
            corr_vis = np.empty_like(resid_vis)

        g, gh, flag_count = self.accumulate_inv_gains(direction=direction)

        self.kernel.compute_corrected(resid_vis, g, gh, corr_vis, 1, 1)

        return corr_vis, flag_count

    def apply_gains(self, vis, full2x2=True):
        """
        Applies the gains to an array at full time-frequency resolution. 

        Args:
            model_arr (np.ndarray):
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing 
                model visibilities.

        Returns:
            np.ndarray:
                Array containing the result of GMG\ :sup:`H`.
        """
        g, gh = self.accumulate_gains()
        self.kernel.apply_gains(vis, g, gh, 1, 1)
        return vis

    def check_convergence(self, min_delta_g):
        """
        Updates the convergence info of the current time-frequency chunk. 

        Args:
            min_delta_g (float):
                Threshold for the minimum change in the gains - convergence criterion.
        """
        return self.active_term.check_convergence(min_delta_g)

    def restrict_solution(self, gains):
        """
        Restricts the solutions by, for example, selecting a reference antenna or taking only the 
        amplitude. 
        """
        return self.active_term.restrict_solution(gains)

    def flag_solutions(self, flags_arr, final=False):
        """ Flags gain solutions."""
        # Per-iteration flagging done on the active term, final (or pre-apply) flagging is done on all terms.
        if final:
            return any([ term.flag_solutions(flags_arr, final) for term in self.jones_terms if term.solvable ])
        else:
            return self.active_term.flag_solutions(flags_arr, 0)

    def num_gain_flags(self, mask=None, final=False):
        if final:
            nfl = [0, 0]
            for term in self.jones_terms:
                n1, n2 = term.num_gain_flags(mask, final=True)
                nfl[0] += int(n1)
                nfl[1] += int(n2)
            return nfl
        else:
            return self.active_term.num_gain_flags(mask)

    def _next_chain_term(self):
        while True:
            if not self.term_iters:
                return False
            previous_term = self.active_term
            # clear converged and stalled properties
            self.active_index = (self.active_index + 1) % self.n_terms
            if self.active_term.solvable:
                self.active_term.maxiter = self.term_iters.pop(0)
                if not self.active_term.maxiter:
                    print("skipping term {}: 0 term iters specified".format(self.active_term.jones_label), file=log(1))
                    continue
                self.active_term.iters = 0
                self._convergence_states_finalized = False
                if previous_term:
                    previous_term.has_converged = previous_term.has_stalled = False
                self.active_term.has_converged = self.active_term.has_stalled = False
                print("activating term {}".format(self.active_term.jones_label), file=log(1))
                return True
            else:
                print("skipping term {}: non-solvable".format(self.active_term.jones_label), file=log(1))


    def next_iteration(self):
        """
        Updates the iteration count on the relevant element of the Jones chain. It will also handle 
        updating the active Jones term. Ultimately, this should handle any complicated 
        convergence/term switching functionality.
        """

        self.last_active_index = self.active_index
        major_step = False

        if self.active_term.has_converged or self.active_term.has_stalled:
            print("term {} {} ({} iters): {}".format(self.active_term.jones_label,
                        "converged" if self.active_term.has_converged else "stalled",
                        self.active_term.iters, self.active_term.final_convergence_status_string), file=log(1))
            self._convergence_states.append(self.active_term.final_convergence_status_string)
            self._convergence_states_finalized = True
            self._next_chain_term()
            major_step = True

        self.active_term.next_iteration()

        return MasterMachine.next_iteration(self)[0], major_step

    def compute_chisq(self, resid_arr, inv_var_chan, require_full=True):
        """Computes chi-square using the active chain term"""
        if require_full:
            return super(JonesChain, self).compute_chisq(resid_arr, inv_var_chan)
        else:
            return self.active_term.compute_chisq(resid_arr, inv_var_chan)

    @property
    def has_valid_solutions(self):
        """Gives corresponding property of the active chain term"""
        return all([term.has_valid_solutions for term in self.jones_terms if term.solvable])

    @property
    def jones_label(self):
        return self.active_term.jones_label

    @property
    def num_converged_solutions(self):
        """Gives corresponding property of the active chain term"""
        return self.active_term.num_converged_solutions

    @property
    def active_term(self):
        if self.active_index is None:
            # setup first solvable term in chain
            self.active_index = -1
            self._next_chain_term()
        return self.jones_terms[self.active_index]

    @property
    def dd_term(self):
        """Gives corresponding property of the active chain term"""
        return any([ term.dd_term for term in self.jones_terms ])

    @property
    def diag_only(self):
        """This property is true if the machine solves using diagonal visibilities only"""
        return all([ term.diag_only for term in self.jones_terms ])

    @property
    def offdiag_only(self):
        """This property is true if the machine solves using off-diagonal visibilities only"""
        return all([ term.offdiag_only for term in self.jones_terms ])

    @property
    def iters(self):
        """Gives corresponding property of the active chain term"""
        return self.active_term.iters

    @iters.setter
    def iters(self, value):
        """Sets corresponding property of the active chain term"""
        self.active_term.iters = value

    @property
    def maxiter(self):
        """Gives corresponding property of the active chain term"""
        return self.active_term.maxiter

    @maxiter.setter
    def maxiter(self, value):
        """Sets corresponding property of the active chain term"""
        self.active_term.maxiter = value

    @property
    def conditioning_status_string(self):
        return "; ".join([term.conditioning_status_string for term in self.jones_terms])

    @property
    def current_convergence_status_string(self):
        """Current status is reported from the active term"""
        return self.active_term.current_convergence_status_string

    @property
    def final_convergence_status_string(self):
        """Final status is reported from all terms"""
        if not self._convergence_states_finalized:
            self._convergence_states.append(self.active_term.final_convergence_status_string)
            self._convergence_states_finalized = True
        return "; ".join(self._convergence_states)

    @property
    def has_converged(self):
        # Chain has converged when term_iters is empty -- since we take off an element each time we converge a term
        return not self.solvable or \
               ( self.active_term.has_converged and not self.term_iters )
        #return np.all([term.has_converged for term in self.jones_terms])

    @property
    def has_stalled(self):
        return self.active_term.has_stalled and not self.term_iters

    @has_stalled.setter
    def has_stalled(self, value):
        self.active_term.has_stalled = value

    @property
    def epsilon(self):
        return self.active_term.epsilon

    @property
    def delta_chi(self):
        return self.active_term.delta_chi


    class Factory(MasterMachine.Factory):
        """
        Note that a ChainMachine Factory expects a list of jones options (one dict per Jones term), not a single dict.
        """
        def __init__(self, machine_cls, grid, double_precision, apply_only, global_options, jones_options):
            # manufacture dict of "Jones options" for the outer chain
            # note that in apply-only mode, we enforce all terms to non-solvable
            opts = dict(label="chain", solvable=not apply_only, sol=global_options['sol'], chain=jones_options)
            if apply_only:
                for term in jones_options:
                    term['solvable'] = False

            self.chain_options = jones_options
            MasterMachine.Factory.__init__(self, machine_cls, grid, double_precision, apply_only,
                                           global_options, opts)

        def init_solutions(self):
            from cubical.main import UserInputError
            for opts in self.chain_options:
                label = opts["label"]
                jones_class = machine_types.get_machine_class(opts['type'])
                if jones_class is None:
                    raise UserInputError("unknown Jones class '{}'".format(opts['type']))
                if not issubclass(jones_class, Complex2x2Gains) and not issubclass(jones_class, ComplexW2x2Gains) and \
                        opts['solvable']:
                    raise UserInputError("only complex-2x2 or robust-2x2 terms can be made solvable in a Jones chain")
                self._init_solutions(label,
                                     self.make_filename(opts["xfer-from"], label) or
                                     self.make_filename(opts["load-from"], label),
                                     bool(opts["xfer-from"]),
                                     self.solvable and opts["solvable"] and self.make_filename(opts["save-to"], label),
                                     jones_class.exportable_solutions(),
                                     dd_term=opts["dd-term"])

        def determine_allocators(self):
            return self.machine_class.determine_allocators(self.jones_options)




# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details

# This module has been adapted from the DDFacet package,
# (c) Cyril Tasse et al., see http://github.com/saopicc/DDFacet
from __future__ import print_function
import os, re, errno
from . import logger
from . import NpShared
from . import shared_dict

log = logger.getLogger("shm_utils")

# store PID here, so that it corresponds to the PID of the parent thread (when we fork off child processes)
_pid = os.getpid()

program_name = "cubical"

def getShmPrefix():
    """Returns prefix used for shared memory arrays. ddf.PID is the convention"""
    return "{}.{}".format(program_name, _pid)

# init SharedDict with the same base name
shared_dict.SharedDict.setBaseName(getShmPrefix())

def getShmName(name, **kw):
    """
    Forms up a name for a shm-backed shared element. This takes the form of "ddf.PID.", where PID is the
    pid of the process where the cache manager was created (so the parent process, presumably), followed
    by a filename of the form "NAME:KEY1_VALUE1:...", as returned by getElementName(). See getElementName()
    for usage.
    """
    # join keyword args into "key=value:key=value:..."
    kws = ":".join([name] + ["%s_%s" % (key, value) for key, value in sorted(kw.items())])
    return "%s.%s" % (getShmPrefix(), kws)

def cleanupShm ():
    """
    Deletes all shared arrays for this process
    """
    NpShared.DelAll(getShmPrefix())
    # above statement don't work for directories and subdirectories
    os.system("rm -rf /dev/shm/%s"%getShmPrefix())

def cleanupStaleShm ():
    """
    Cleans up "stale" shared memory from previous runs of CubeCal
    """
    # check for stale shared memory
    uid = os.getuid()
    # list of all files in /dev/shm/ matching ddf.PID.* and belonging to us
    shmlist = [ ("/dev/shm/"+filename, re.match('(sem\.)?' + program_name + '\.([0-9]+)(\..*)?$',filename)) for filename in os.listdir("/dev/shm/")
                if os.stat("/dev/shm/"+filename).st_uid == uid ]
    # convert to list of filename,pid tuples
    shmlist = [ (filename, int(match.group(2))) for filename, match in shmlist if match ]
    # now check all PIDs to find dead ones
    # if we get ESRC error from sending signal 0 to the process, it's not running, so we mark it as dead
    dead_pids = set()
    for pid in set([x[1] for x in shmlist]):
        try:
            os.kill(pid, 0)
        except OSError as err:
            if err.errno == errno.ESRCH:
                dead_pids.add(pid)
    # ok, make list of candidates for deletion
    victims = [ filename for filename,pid in shmlist if pid in dead_pids ]
    if victims:
        print("reaping %d shared memory objects associated with %d dead cubical processes"%(len(victims), len(dead_pids)), file=log)
        dirs = [ v for v in victims if os.path.isdir(v) ]
        files = [ v for v in victims if not os.path.isdir(v) ]
        # rm -fr only works for a limited number of arguments (which the semaphore list can easily exceed)
        # so use os.unlink() to remove files, and rm -fr for directories
        for path in files:
            os.unlink(path)
        os.system("rm -fr " + " ".join(dirs))
        # print "rm -fr " + " ".join(victims)




import six
import numpy as np

if six.PY3:
    long = int

def assert_isint(v):
    if isinstance(v, (int, long, bool)):
        return True
    elif isinstance(v, list):
        return all(map(lambda x: isinstance(x, (int, long, bool)), v))
    elif isinstance(v, np.ndarray):
        return v.dtype in [np.int, np.int16, np.int32, np.int64, np.int8,
                           np.int_, np.intc, np.integer, np.bool, int, long]
    else:
        return False

def assert_isfp(v):
    if isinstance(v, float):
        return True
    elif isinstance(v, list):
        return all(map(lambda x: isinstance(x, float), v))
    elif isinstance(v, np.ndarray):
        return v.dtype in [np.float, np.float_, np.float64, np.float16, np.float32, np.double]
    else:
        return False